"""
Description
===========

Module: ``geocompy.communication``

Implementations of connection methods.

Functions
---------

- ``get_logger``
- ``open_serial``

Types
-----

- ``Connection``
- ``SerialConnection``
"""
from __future__ import annotations

import logging
from types import TracebackType
from typing import Literal, Generator
from contextlib import contextmanager
from abc import ABC, abstractmethod
from time import sleep

from serial import (
    Serial,
    SerialException,
    SerialTimeoutException,
    PARITY_NONE
)


def get_logger(
    name: str,
    target: Literal['null', 'file', 'stdout'] = 'null',
    level: int = logging.NOTSET,
    file: str = ""
) -> logging.Logger:
    """
    Utility function that creates a logger instance with standard
    formatting, logging to the specified target.

    Parameters
    ----------
    name : str
        Name of the logger.
    target : Literal['null', 'file', 'stdout'], optional
        Logging target, by default 'null'
    level : int, optional
        Logging level, by default logging.NOTSET
    file : str, optional
        Path to target log file (**must not be** ``""`` when target is
        'file'), by default ""

    Returns
    -------
    logging.Logger

    Note
    ----
    If a logger with the specified name already exists, it will be
    overwritten with the newly created handlers.
    """
    log = logging.getLogger(name)
    log.handlers.clear()
    log.setLevel(level)
    fmt = logging.Formatter(
        "%(asctime)s <%(name)s> [%(levelname)s] %(message)s",
        "%Y-%m-%d %H:%M:%S"
    )
    match target:
        case "null":
            log.addHandler(logging.NullHandler())
        case "file" if file != "":
            fhandler = logging.FileHandler(
                file,
                encoding="utf8"
            )
            fhandler.setFormatter(fmt)
            log.addHandler(fhandler)
        case "stdout":
            shandler = logging.StreamHandler()
            shandler.setFormatter(fmt)
            log.addHandler(shandler)

    return log


class Connection(ABC):
    """
    Interface definition for connection implementations.
    """

    @abstractmethod
    def is_open(self) -> bool: ...

    @abstractmethod
    def send(self, message: str) -> None: ...

    @abstractmethod
    def receive(self) -> str: ...

    @abstractmethod
    def exchange(self, cmd: str) -> str: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def reset(self) -> None: ...


def open_serial(
    port: str,
    *,
    speed: int = 9600,
    databits: int = 8,
    stopbits: int = 1,
    parity: str = PARITY_NONE,
    timeout: int = 15,
    eom: str = "\r\n",
    eoa: str = "\r\n",
    sync_after_timeout: bool = False,
    retry: int = 1
) -> SerialConnection:
    """
    Constructs a SerialConnection with the given communication
    parameters.

    Parameters
    ----------
    port : str
        Name of the port to use (e.g. ``COM1`` or ``/dev/ttyUSB0``).
    speed : int, optional
        Communication speed (baud), by default 9600
    databits : int, optional
        Number of data bits, by default 8
    stopbits : int, optional
        Number of stop bits, by default 1
    parity : str, optional
        Parity bit behavior, by default PARITY_NONE
    timeout : int, optional
        Communication timeout threshold, by default 15
    eom : str, optional
        EndOfMessage sequence, by default ``"\\r\\n"``
    eoa : str, optional
        EndOfAnswer sequence, by default ``"\\r\\n"``
    sync_after_timeout : bool, optional
        Attempt to re-sync the message-response que, if a timeout
        occured in the previous exchange, by default False
    retry : int, optional
        Number of retry attempts if the connection opening fails, by
        default 1

    Returns
    -------
    SerialConnection

    Warning
    -------

    The syncing feature should be used with caution! See `SerialConnection`
    for more information!

    Examples
    --------

    Opening a serial connection similar to a file:

    >>> conn = open_serial("COM1", speed=19200, timeout=5)
    >>> # do operations
    >>> conn.close()

    Using as a context manager:

    >>> with open_serial("COM1", timeout=20) as conn:
    ...     conn.send("test")

    """
    exc: Exception = Exception()
    for i in range(retry):
        try:
            serialport = Serial(
                port, speed, databits, parity, stopbits, timeout
            )
            break
        except Exception as e:
            exc = e

        sleep(5)
    else:
        raise exc

    wrapper = SerialConnection(
        serialport,
        eom=eom,
        eoa=eoa,
        sync_after_timeout=sync_after_timeout
    )
    return wrapper


class SerialConnection(Connection):
    """
    Connection wrapping an open serial port.

    The passed serial port should be configured and opened in advance.
    Take care to set the approriate speed (baud), data bits, timeout etc.
    For most instruments a 9600 speed setting, 8 data + 1 stop bits is
    correct. A suitable timeout for total stations might be 15 seconds.
    (A too short timeout may result in unexpected errors when waiting for
    a slower, motorized function.)

    Examples
    --------

    Setting up a basic serial connection:

    >>> from serial import Serial
    >>> port = Serial("COM4", timeout=15)
    >>> conn = gc.communication.SerialConnection(port)
    >>> # message exchanges
    >>> conn.close()

    Using as a context manager:

    >>> from serial import Serial
    >>> port = Serial("COM4", timeout=15)
    >>> with gc.communication.SerialConnection(port) as conn:
    ...     # message exchanges
    >>>
    >>> port.is_open
    False
    >>> # port is automatically closed when the context is exited

    """

    def __init__(
        self,
        port: Serial,
        *,
        eom: str = "\r\n",
        eoa: str = "\r\n",
        sync_after_timeout: bool = False
    ):
        """
        Parameters
        ----------
        port : Serial
            Serial port to communicate on.
        eom : str, optional
            EndOfMessage sequence, by default ``"\\r\\n"``
        eoa : str, optional
            EndOfAnswer sequence, by default ``"\\r\\n"``
        sync_after_timeout : bool, optional
            Attempt to re-sync the message-response que, if a timeout
            occured in the previous exchange, by default False

        Notes
        -----
        If the serial port is not already open, the opening will be
        attempted. This might raise an exception if the port cannot
        be opened.

        Warning
        -------

        The que syncing is attempted by repeatedly reading from the
        receiving buffer, as many times as a timeout was previously
        detected. This can only solve issues, if the connection target
        was just slow, and not completely unresponsive. If the target
        became truly unresponsive, but came back online later, the sync
        attempt can cause further problems. Use with caution!

        (Timeouts should be avoided when possible, use a temporary override
        on exchanges that are expected to finish in a longer time.)

        """

        self._port: Serial = port
        self.eom: str = eom  # end of message
        self.eoa: str = eoa  # end of answer
        self._attempt_sync: bool = sync_after_timeout
        self._timeout_counter: int = 0

        if not self._port.is_open:
            self._port.open()

    def __del__(self) -> None:
        self._port.close()

    def __enter__(self) -> SerialConnection:
        return self

    def __exit__(
        self,
        exc_type: BaseException,
        exc_value: BaseException,
        exc_tb: TracebackType
    ) -> None:
        self._port.close()

    def close(self) -> None:
        """
        Closes the serial port.
        """
        self._port.close()

    def is_open(self) -> bool:
        """
        Checks if the serial port is currently open.

        Returns
        -------
        bool
            State of the port.

        """
        return self._port.is_open

    def send(self, message: str) -> None:
        """
        Writes a single message to the serial line.

        Parameters
        ----------
        message : str
            Message to send.

        Raises
        ------
        ~serial.SerialException
            If the serial port is not open.

        """
        if not self._port.is_open:
            raise SerialException(
                "serial port is not open"
            )

        if not message.endswith(self.eom):
            message += self.eom

        payload = message.encode('ascii', 'ignore')
        self._port.write(payload)

    def receive(self) -> str:
        """
        Reads a single message from the serial line.

        Returns
        -------
        str
            Received message.

        Raises
        ------
        ~serial.SerialException
            If the serial port is not open.
        ~serial.SerialTimeoutException
            If the connection timed out before receiving the
            EndOfAnswer sequence.

        """
        if not self._port.is_open:
            raise SerialException(
                "serial port is not open"
            )

        eoabytes = self.eoa.encode("ascii")
        if self._attempt_sync and self._timeout_counter > 0:
            for i in range(self._timeout_counter):
                excess = self._port.read_until(eoabytes)
                if not excess.endswith(eoabytes):
                    self._timeout_counter += 1
                    raise SerialTimeoutException(
                        "Serial connection timed on 'receive' during"
                        "attempt to recover from a previous timeout"
                    )
            else:
                self._timeout_counter = 0

        answer = self._port.read_until(eoabytes)
        if not answer.endswith(eoabytes):
            self._timeout_counter += 1
            raise SerialTimeoutException(
                "serial connection timed out on 'receive'"
            )

        return answer.decode("ascii").removesuffix(self.eoa)

    def exchange(self, cmd: str) -> str:
        """
        Writes a message to the serial line, and receives the
        corresponding response.

        Parameters
        ----------
        cmd : str
            Message to send.

        Returns
        -------
        str
            Response to the sent message

        Raises
        ------
        ~serial.SerialException
            If the serial port is not open.
        ~serial.SerialTimeoutException
            If the connection timed out before receiving the
            EndOfAnswer sequence for one of the responses.

        """
        self.send(cmd)
        return self.receive()

    def reset(self) -> None:
        """
        Resets the connection by clearing the incoming and outgoing
        buffers, and resetting the internal state. This could be used
        to recover from a desync (possibly caused by a timeout).

        Warning
        -------

        Trying to recover after repeated timeouts with a hard reset can
        cause further issues, if the reset is attempted while responses
        were finally being received. It is recommended to wait some time
        after the last command was sent, before resetting.
        """
        self._port.reset_input_buffer()
        self._port.reset_output_buffer()
        self._timeout_counter = 0

    @contextmanager
    def timeout_override(
        self,
        timeout: int | None
    ) -> Generator[None, None, None]:
        """
        Context manager that temporarily overrides connection parameters.

        Parameters
        ----------
        timeout : int | None
            Temporary timeout in seconds. Set to None to wait indefinitely.

        Returns
        -------
        Generator
            Context manager generator object.

        Warning
        -------
        An indefinite timeout might leave the connection in a perpetual
        waiting state, if the instrument became unresponsive in the
        mean time (e.g. it powered off due to low battery charge).

        Example
        -------

        >>> from serial import Serial
        >>> from geocompy.communication import SerialConnection
        >>>
        >>> port = Serial("COM1", timeout=5)
        >>> with SerialConnection(port) as com:
        ...     # normal operation
        ...
        ...     # potentially long operation
        ...     with com.timeout_override(20):
        ...         answer = com.exchange("message")
        ...
        ...     # continue normal operation
        ...
        """
        saved_timeout = self._port.timeout

        try:
            self._port.timeout = timeout
            yield
        finally:
            self._port.timeout = saved_timeout



"""
Description
===========

Module: ``geocompy.geo``

The ``geo`` package provides wrapper methods for all GeoCom RPC functions
available on supported instruments running various versions of the TPS
system software, or software based on the TPS system.

Types
-----

- ``GeoCom``

Submodules
----------

- ``geocompy.geo.gcdata``
- ``geocompy.geo.gctypes``
- ``geocompy.geo.aus``
- ``geocompy.geo.aut``
- ``geocompy.geo.bap``
- ``geocompy.geo.bmm``
- ``geocompy.geo.cam``
- ``geocompy.geo.com``
- ``geocompy.geo.csv``
- ``geocompy.geo.ctl``
- ``geocompy.geo.dna``
- ``geocompy.geo.edm``
- ``geocompy.geo.ftr``
- ``geocompy.geo.img``
- ``geocompy.geo.kdm``
- ``geocompy.geo.mot``
- ``geocompy.geo.sup``
- ``geocompy.geo.tmc``
- ``geocompy.geo.wir``
"""
import re
from logging import Logger
from traceback import format_exc
from time import sleep
from typing import Any, Callable, Iterable, overload, TypeVar

from serial import SerialException, SerialTimeoutException

from geocompy.data import Angle, Byte
from geocompy.communication import Connection, get_logger
from .gctypes import (
    GeoComCode,
    GeoComResponse,
    GeoComType,
    rpcnames
)
from .aus import GeoComAUS
from .aut import GeoComAUT
from .bap import GeoComBAP
from .bmm import GeoComBMM
from .cam import GeoComCAM
from .com import GeoComCOM
from .csv import GeoComCSV
from .ctl import GeoComCTL
from .dna import GeoComDNA
from .edm import GeoComEDM
from .ftr import GeoComFTR
from .img import GeoComIMG
from .kdm import GeoComKDM
from .mot import GeoComMOT
from .sup import GeoComSUP
from .tmc import GeoComTMC
from .wir import GeoComWIR


_T = TypeVar("_T")
_MAX_TRANSACTION = 2**15


class GeoCom(GeoComType):
    """
    GeoCom protocol handler.

    The individual commands are available through their respective
    subsystems.

    Examples
    --------

    Opening a simple serial connection:

    >>> from geocompy.communication import open_serial
    >>> from geocompy.geo import GeoCom
    >>>
    >>> with open_serial("COM1") as line:
    ...     tps = GeoCom(line)
    ...     tps.com.nullprocess()
    ...
    >>>

    Passing a logger:

    >>> from logging import DEBUG
    >>> from geocompy.communication import open_serial, get_logger
    >>> from geocompy.geo import GeoCom
    >>>
    >>> log = get_logger("Viva", "stdout", DEBUG)
    >>> with open_serial("COM1") as line:
    ...     tps = GeoCom(line, log)
    ...     tps.com.nullprocess()
    ...
    >>>
    GeoComResponse(COM_NullProc) ... # Startup connection test
    GeoComResponse(COM_GetDoublePrecision) ... # Precision sync
    GeoComResponse(COM_NullProc) ... # First executed command
    """
    _R1P: re.Pattern[str] = re.compile(
        r"^%R1P,"
        r"(?P<comrc>\d+),"
        r"(?P<tr>\d+):"
        r"(?P<rc>\d+)"
        r"(?:,(?P<params>.*))?$"
    )

    def __init__(
        self,
        connection: Connection,
        logger: Logger | None = None,
        retry: int = 2
    ):
        """
        After the subsystems are initialized, the connection is tested by
        sending an ``LF`` character to clear the receiver buffer, then the
        ``COM_NullProc`` is executed. If the test fails, it is retried with
        one second delay. The test is attempted `retry` number of times.

        Parameters
        ----------
        connection : Connection
            Connection to use for communication
            (usually a serial connection).
        logger : Logger | None, optional
            Logger to log all requests and responses, by default None
        retry : int, optional
            Number of retries at connection validation before giving up.

        Raises
        ------
        ConnectionError
            If the connection could not be verified in the specified
            number of retries.
        """
        self.transaction_counter = 0
        """Number of command transactions started during the current
        session."""
        self._conn: Connection = connection
        if logger is None:
            logger = get_logger("/dev/null")
        self._logger: Logger = logger
        self._precision = 15

        self.aus: GeoComAUS = GeoComAUS(self)
        """
        Alt User subsystem.

        .. versionadded:: GeoCom-TPS1100-1.04
        """
        self.aut: GeoComAUT = GeoComAUT(self)
        """Automation subsystem."""
        self.bap: GeoComBAP = GeoComBAP(self)
        """Basic applications subsystem."""
        self.bmm: GeoComBMM = GeoComBMM(self)
        """Basic man-machine interface subsystem."""
        self.cam: GeoComCAM = GeoComCAM(self)
        """
        Camera subsystem.

        .. versionadded:: GeoCom-VivaTPS
        """
        self.com: GeoComCOM = GeoComCOM(self)
        """Communications subsystem."""
        self.csv: GeoComCSV = GeoComCSV(self)
        """Central services subsystem."""
        self.ctl: GeoComCTL = GeoComCTL(self)
        """
        Control task subsystem.

        .. versionremoved:: GeoCom-TPS1200
        """
        self.dna: GeoComDNA = GeoComDNA(self)
        """
        Digital level subsystem.

        .. versionadded:: GeoCom-LS
        """
        self.edm: GeoComEDM = GeoComEDM(self)
        """Electronic distance measurement subsystem."""
        self.ftr: GeoComFTR = GeoComFTR(self)
        """
        File transfer subsystem.

        .. versionadded:: GeoCom-TPS1200
        """
        self.img: GeoComIMG = GeoComIMG(self)
        """
        Image processing subsystem.

        .. versionadded:: GeoCom-TPS1200
        """
        self.kdm: GeoComKDM = GeoComKDM(self)
        """
        Keyboard display unit subsystem.

        .. versionadded:: GeoCom-VivaTPS
        """
        self.mot: GeoComMOT = GeoComMOT(self)
        """Motorization subsytem."""
        self.sup: GeoComSUP = GeoComSUP(self)
        """Supervisor subsystem."""
        self.tmc: GeoComTMC = GeoComTMC(self)
        """Theodolite measurement and calculation subsystem."""
        self.wir: GeoComWIR = GeoComWIR(self)
        """
        Word Index registration subsystem.

        .. versionremoved:: GeoCom-TPS1200
        """

        for i in range(retry):
            try:
                self._conn.send("\n")
                if self.com.nullprocess():
                    sleep(1)
                    break
            except Exception:
                self._logger.exception("Exception during connection attempt")

            sleep(1)
        else:
            raise ConnectionError(
                "could not establish connection to instrument"
            )

        resp = self.get_double_precision()
        if resp.params is not None:
            self._precision = resp.params
            self._logger.info(f"Synced double precision: {self._precision}")
        else:
            self._logger.error(
                f"Could not syncronize double precision, "
                f"defaulting to {self._precision:d}"
            )

        self._logger.info("Connection initialized")
        name = self.csv.get_instrument_name().params or "Unknown"
        geocom = self.com.get_geocom_version().params or (0, 0, 0)
        firmware = self.csv.get_firmware_version().params or (0, 0, 0)
        self._logger.info(
            f"Instrument: {name} "
            f"(firmware: v{firmware[0]}.{firmware[1]}.{firmware[2]}, "
            f"geocom: v{geocom[0]}.{geocom[1]}.{geocom[2]})"
        )

    @overload
    def request(
        self,
        rpc: int,
        params: Iterable[int | float | bool | str | Angle | Byte] = (),
        parsers: Callable[[str], _T] | None = None
    ) -> GeoComResponse[_T]: ...

    @overload
    def request(
        self,
        rpc: int,
        params: Iterable[int | float | bool | str | Angle | Byte] = (),
        parsers: Iterable[Callable[[str], Any]] | None = None
    ) -> GeoComResponse[tuple[Any, ...]]: ...

    def request(
        self,
        rpc: int,
        params: Iterable[int | float | bool | str | Angle | Byte] = (),
        parsers: (
            Iterable[Callable[[str], Any]]
            | Callable[[str], Any]
            | None
        ) = None
    ) -> GeoComResponse[Any]:
        """
        Executes an RPC request and returns the parsed GeoCom response.

        Constructs a request(from the given RPC code and parameters),
        writes it to the serial line, then reads the response. The
        response is then parsed using the provided parser functions.

        Parameters
        ----------
        rpc: int
            Number of the RPC to execute.
        params: Iterable[int | float | bool | str | Angle | Byte]
            Parameters for the request, by default()
        parsers: Iterable[Callable[[str], Any]] \
                  | Callable[[str], Any] \
                  | None, optional
            Parser functions for the values in the RPC response,
            by default None

        Returns
        -------
        GeoComResponse
            Parsed return codes and parameters from the RPC response.

        """
        strparams: list[str] = []
        for item in params:
            match item:
                case Angle():
                    value = f"{round(float(item), self._precision):f}"
                    value = value.rstrip("0")
                    if value[-1] == ".":
                        value += "0"
                case Byte():
                    value = str(item)
                case float():
                    value = f"{round(item, self._precision):f}".rstrip("0")
                    if value[-1] == ".":
                        value += "0"
                case int():
                    value = f"{item:d}"
                case str():
                    value = f"\"{item}\""
                case _:
                    raise TypeError(f"unexpected parameter type: {type(item)}")

            strparams.append(value)

        trid = self.transaction_counter % _MAX_TRANSACTION
        self.transaction_counter += 1
        cmd = f"%R1Q,{rpc},{trid}:{','.join(strparams)}"
        try:
            answer = self._conn.exchange(cmd)
        except SerialTimeoutException:
            self._logger.error(format_exc())
            answer = (
                f"%R1P,{GeoComCode.COM_TIMEDOUT:d},"
                f"{trid}:{GeoComCode.OK:d}"
            )
        except SerialException:
            self._logger.error(format_exc())
            answer = (
                f"%R1P,{GeoComCode.COM_CANT_SEND:d},"
                f"{trid}:{GeoComCode.OK:d}"
            )
        except Exception:
            self._logger.error(format_exc())
            answer = (
                f"%R1P,{GeoComCode.COM_FAILED:d},"
                f"{trid}:{GeoComCode.OK:d}"
            )

        response = self.parse_response(
            cmd,
            answer,
            parsers
        )
        self._logger.debug(response)
        return response

    @overload
    def parse_response(
        self,
        cmd: str,
        response: str,
        parsers: Callable[[str], _T] | None = None
    ) -> GeoComResponse[_T]: ...

    @overload
    def parse_response(
        self,
        cmd: str,
        response: str,
        parsers: Iterable[Callable[[str], Any]] | None = None
    ) -> GeoComResponse[tuple[Any, ...]]: ...

    def parse_response(
        self,
        cmd: str,
        response: str,
        parsers: (
            Iterable[Callable[[str], Any]]
            | Callable[[str], Any]
            | None
        ) = None
    ) -> GeoComResponse[Any]:
        """
        Parses RPC response and constructs `GeoComResponse` instance.

        Parameters
        ----------
        cmd: str
            Full, serialized request, that invoked the response.
        response: str
            Full, received response.
        parsers: Iterable[Callable[[str], Any]] \
                  | Callable[[str], Any] \
                  | None, optional
            Parser functions for the values in the RPC response,
            by default None

        Returns
        -------
        GeoComResponse
            Parsed return codes and parameters from the RPC response.

        """
        m = self._R1P.match(response)
        rpc = int(cmd.split(":")[0].split(",")[1])
        trid_expected = int(cmd.split(":")[0].split(",")[2])
        rpcname = rpcnames.get(rpc, str(rpc))
        if not m:
            return GeoComResponse(
                rpcname,
                cmd,
                response,
                GeoComCode.COM_CANT_DECODE,
                GeoComCode.OK,
                0
            )

        groups = m.groupdict()
        trid = int(groups.get("tr", "-1"))
        if trid != trid_expected:
            return GeoComResponse(
                rpcname,
                cmd,
                response,
                GeoComCode.COM_TR_ID_MISMATCH,
                GeoComCode.OK,
                trid
            )

        try:
            comrc = GeoComCode(int(groups["comrc"]))
        except Exception:
            comrc = GeoComCode.UNDEFINED

        try:
            rc = GeoComCode(int(groups["rc"]))
        except Exception:
            rc = GeoComCode.UNDEFINED

        values = groups.get("params", "")
        if values is None:
            values = ""

        if values == "":
            return GeoComResponse(
                rpcname,
                cmd,
                response,
                comrc,
                rc,
                trid
            )

        if parsers is None:
            parsers = ()
        elif not isinstance(parsers, Iterable):
            parsers = (parsers,)

        params: list[Any] = []
        try:
            for func, value in zip(parsers, values.split(",")):
                params.append(func(value))
        except Exception:
            return GeoComResponse(
                rpcname,
                cmd,
                response,
                GeoComCode.COM_CANT_DECODE,
                GeoComCode.OK,
                0
            )

        match len(params):
            case 0:
                params_final = None
            case 1:
                params_final = params[0]
            case _:
                params_final = tuple(params)

        return GeoComResponse(
            rpcname,
            cmd,
            response,
            comrc,
            rc,
            int(groups["tr"]),
            params_final
        )

    def get_double_precision(self) -> GeoComResponse[int]:
        """
        RPC 108, ``COM_GetDoublePrecision``

        Gets the current ASCII communication floating point precision of
        the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Floating point decimal places.

        See Also
        --------
        set_double_precision
        """
        return self.request(
            108,
            parsers=int
        )

    def set_double_precision(
        self,
        digits: int
    ) -> GeoComResponse[None]:
        """
        RPC 107, ``COM_SetDoublePrecision``

        Sets the ASCII communication floating point precision of the
        instrument.

        Parameters
        ----------
        digits: int
            Floating points decimal places.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_double_precision
        """
        response: GeoComResponse[None] = self.request(107, [digits])
        if not response.error:
            self._precision = digits
        return response

    def abort(self) -> GeoComResponse[None]:
        """
        Aborts the current motorized function.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``UNDEFINED``: The command was accepted, but the response
                  indicated failed execution. Possibly nothing to abort.
                - ``COM_FAILED``: The abort command could not be sent due
                  to a communication issue.

        Note
        ----
        The abort command does not actually use the GeoCom syntax, but the
        function returns a `GeoComResponse` anyway, in order to maintain
        the uniformity with the rest of the commands.
        """
        cmd = "c"
        try:
            ans = self._conn.exchange(cmd)
            rpccode = GeoComCode.OK if ans == "?" else GeoComCode.UNDEFINED
            comcode = GeoComCode.OK
        except Exception:
            ans = ""
            rpccode = GeoComCode.OK
            comcode = GeoComCode.COM_FAILED

        return GeoComResponse(
            "Abort",
            cmd,
            ans,
            comcode,
            rpccode,
            0
        )



"""
GeoComPy
========

Python wrapper functions for communicating with surveying
instruments over a serial connection.

The implementations use the Leica GeoCom ASCII RPC procotol primarily.
For older instruments, that do not support it, the GSI Online commands
are used instead.

The package provides
    1. Utility data types for handling instrument responses
    2. Instrument software specific low level commands

Documentation
-------------

Public classes and methods are provided with proper docstrings, that can
be viewed in the source code, through introspection tools or editor
utilities. The docstrings follow the NumPy style conventions. In addition
to the in-code documentation, a complete, rendered reference is avaialable
on the `GeoComPy documentation <https://geocompy.readthedocs.io>`_ site.

Some docstrings provide examples. These examples assume that `geocompy`
has been imported as ``gc``:

    >>> import geocompy as gc

Subpackages
-----------

``geocompy.geo``
    Communication through GeoCom protocol.

``geocompy.gsi``
    Communication through GSI Online protocol.

Submodules
----------

``geocompy.data``
    Utilities for data handling.

``geocompy.communication``
    Communication methods.

Reexports
---------

``geocompy.data.Angle``
    Angle value primitive.

``geocompy.data.Vector``
    3D vector primitive.

``geocompy.data.Coordinate``
    3D coordinate primitive.

``geocompy.communication.open_serial``
    Serial connection context manager function.

``geocompy.communication.get_logger``
    Utility function to create logger objects.

``geocompy.gsi.dna.GsiOnlineDNA``
    DNA instrument implementation.

``geocompy.gsi.gsitypes.GsiOnlineResponse``
    GSI Online protocol response container.

``geocompy.geo.GeoCom``
    GeoCom protocol handler.

``geocompy.geo.gctypes.GeoComCode``
    GeoCom return codes.

``geocompy.geo.gctypes.GeoComResponse``
    GeoCom protocol response container.
"""
try:
    from ._version import __version__ as __version__
except Exception:
    __version__ = "0.0.0"  # Placeholder value for source installs

from .data import (  # noqa: F401
    Angle as Angle,
    Vector as Vector,
    Coordinate as Coordinate
)

from .communication import (  # noqa: F401
    open_serial as open_serial,
    get_logger as get_logger
)

from .gsi.gsitypes import GsiOnlineResponse as GsiOnlineResponse  # noqa: F401
from .gsi.dna import GsiOnlineDNA as GsiOnlineDNA  # noqa: F401

from .geo.gctypes import GeoComResponse as GeoComResponse  # noqa: F401
from .geo.gctypes import GeoComCode as GeoComCode  # noqa: F401
from .geo import GeoCom as GeoCom  # noqa: F401



"""
Description
===========

Module: ``geocompy.gsi.gsidata``

The GSI Online data module provides utility types, that are specific to
the GSI Online protocol.

Enums
-----

- ``AutoOff``
- ``BeepIntensity``
- ``Communication``
- ``GSIFormat``
- ``Illumination``
- ``RecordCode``
- ``Recorder``
- ``Units``
"""
from enum import Enum


class Communication:
    """GSI Online serial communication settings."""
    class Baud(Enum):
        """Serial speed."""
        B1200 = 2
        B2400 = 3
        B4800 = 4
        B9600 = 5
        B19200 = 6

    class Parity(Enum):
        """Serial parity."""
        NONE = 0
        ODD = 1
        EVEN = 2

    class Terminator(Enum):
        """Message terminator."""
        CR = 0
        CRLF = 1


class Units:
    """GSI Online measurement units."""
    class Temperature(Enum):
        """Temperature unit."""
        CELSIUS = 0
        FAHRENHEIT = 1

    class Distance(Enum):
        """Distance unit."""
        METER = 0
        USFEET = 1
        FEET = 2
        USFEETINCH = 5


class BeepIntensity(Enum):
    """GSI Online beep signal intensitie."""
    OFF = 0
    MEDIUM = 1
    LOUD = 2


class Illumination(Enum):
    """GSI Online illumination setting."""
    OFF = 0
    LEVELONLY = 2
    BOTH = 3


class Recorder(Enum):
    """GSI Online data recorder device."""
    INTERNAL = 0
    RS232 = 1


class AutoOff(Enum):
    """GSI Online automatic shutdown."""
    OFF = 0
    ON = 1
    SLEEP = 2


class RecordCode(Enum):
    """GSI Online automatic code recording."""
    BEFORE = 0
    """Record code before measurement."""
    AFTER = 1
    """Record code after measurement."""


class GSIFormat(Enum):
    """GSI Online recording format."""
    GSI8 = 0
    GSI16 = 1



"""
Description
===========

Module: ``geocompy.geo.ftr``

Definitions for the GeoCom File transfer subsystem.

Types
-----

- ``GeoComPFTR``

"""
from __future__ import annotations

from datetime import datetime

from ..data import (
    Byte,
    parsestr,
    parsebool,
    toenum
)
from .gcdata import (
    Device,
    File
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComFTR(GeoComSubsystem):
    """
    File transfer subsystem of the GeoCom protocol.

    This subsystem provides access to the internal file system of the
    instrument, and provides methods to list or download files.

    .. versionadded:: GeoCom-TPS1200
    """

    def setup_listing(
        self,
        device: Device | str = Device.CFCARD,
        filetype: File | str = File.UNKNOWN,
        path: str = ""
    ) -> GeoComResponse[None]:
        """
        RPC 23306, ``FTR_SetupList``

        Prepares file listing of the specified file type, on the selected
        memory device.

        Parameters
        ----------
        device : Device | str, optional
            Memory device, by default CFCARD
        filetype : File | str, optional
            File type, by default UNKNOWN
        path : str, optional
            Search path, by default ""

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Memory device unavailable, or cannot find
                  search path.
                - ``NOTOK``: Setup already exists, previous setup was not
                  aborted.
                - ``FTR_FILEACCESS``: File access error.

        See Also
        --------
        list
        abort_list

        """
        _device = toenum(Device, device)
        _filetype = toenum(File, filetype)
        return self._request(
            23306,
            [_device.value, _filetype.value, path]
        )

    def list(
        self,
        next: bool = False
    ) -> GeoComResponse[tuple[bool, str, int, datetime | None]]:
        """
        RPC 23307, ``FTR_List``

        Gets a single file entry according to the predefined listing
        setup.

        Parameters
        ----------
        next : bool, optional
            Get the next item, after a previous call (get first item
            otherwise), by default False

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: If file is last in listing.
                - `str`: Name of the file.
                - `int`: File size [byte].
                - `datetime` | None: Date and time of last modification.
            Error codes:
                - ``FTR_MISSINGSETUP``: No active listing setup.
                - ``FTR_INVALIDINPUT``: First item is missing, or last
                  call was already the last.

        See Also
        --------
        setup_listing
        abort_list

        """
        def transform(
            params: tuple[
                bool, str, int,
                Byte, Byte, Byte, Byte, Byte, Byte, Byte
            ] | None
        ) -> tuple[bool, str, int, datetime | None] | None:
            if params is None:
                return None

            return (
                params[0],
                params[1],
                params[2],
                datetime(
                    int(params[9]) + 2000,
                    int(params[8]),
                    int(params[7]),
                    int(params[3]),
                    int(params[4]),
                    int(params[5]),
                    int(params[6]) * 10000
                ) if params[1] != "" else None
            )

        response = self._request(
            23307,
            [next],
            (
                parsebool,
                parsestr,
                int,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse
            )
        )
        return response.map_params(transform)

    def abort_list(self) -> GeoComResponse[None]:
        """
        RPC 23308, ``FTR_AbortList``

        Aborts current file listing setup.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        setup_listing
        list

        """
        return self._request(23308)

    def setup_download(
        self,
        filename: str,
        blocksize: int,
        device: Device | str = Device.CFCARD,
        filetype: File | str = File.UNKNOWN,
    ) -> GeoComResponse[int]:
        """
        RPC 23303, ``FTR_SetupDownload``

        Prepares download of the specified file of the specified type, on
        the selected memory device.

        Parameters
        ----------
        filename : str
            File name (or full path if type is unknown).
        blocksize : int
            Download data block size.
        device : Device | str, optional
            Memory device, by default CFCARD
        filetype : File | str, optional
            File type, by default UNKNOWN

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Number of download blocks needed.
            Error codes:
                - ``IVPARAM``: Memory device unavailable, or cannot find
                  file path.
                - ``NOTOK``: Setup already exists, previous setup was not
                  aborted.
                - ``FTR_INVALIDINPUT``: Block size too big.
                - ``FTR_FILEACCESS``: File access error.

        See Also
        --------
        download
        abort_download

        """
        _device = toenum(Device, device)
        _filetype = toenum(File, filetype)
        return self._request(
            23303,
            [_device.value, _filetype.value, filename, blocksize],
            int
        )

    def download(
        self,
        block: int
    ) -> GeoComResponse[str]:
        """
        RPC 23304, ``FTR_Download``

        Downloads a single data block of a previously defined download
        sequence.

        Parameters
        ----------
        block : int
            Number of data block to download.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Data block as hexadecimal byte string.
            Error codes:
                - ``FTR_MISSINGSETUP``: No active download setup.
                - ``FTR_INVALIDINPUT``: First block is missing.
                - ``FTR_FILEACCESS``: File access error.

        See Also
        --------
        setup_download
        abort_download

        """
        return self._request(
            23304,
            [block],
            parsestr
        )

    def abort_download(self) -> GeoComResponse[None]:
        """
        RPC 23305, ``FTR_AbortDownload``

        Aborts current file download setup.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        setup_download
        download

        """
        return self._request(23305)

    def delete(
        self,
        filename: str,
        time: datetime | None = None,
        device: Device | str = Device.CFCARD,
        filetype: File | str = File.UNKNOWN
    ) -> GeoComResponse[int]:
        """
        RPC 23309, ``FTR_Delete``

        Deletes one or more files. Wildcards can be used to delete
        multiple items. If a date is given, only files older than that
        date are deleted.

        Parameters
        ----------
        filename : str
            File name (or full path if type is unknown).
        time : datetime | None, optional
            Deletion limit date, by default None
        device : Device | str, optional
            Memory device, by default CFCARD
        filetype : File | str, optional
            File type, by default UNKNOWN

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Number of files deleted.
            Error codes:
                - ``IVPARAM``: Memory device unavailable, or cannot find
                  file path.

        See Also
        --------
        list

        """
        _device = toenum(Device, device)
        _filetype = toenum(File, filetype)
        if time is None:
            params = [
                _device.value, _filetype.value,
                Byte(0), Byte(0), Byte(0),
                filename
            ]
        else:
            params = [
                _device.value, _filetype.value,
                Byte(time.day), Byte(time.month), Byte(time.year - 2000),
                filename
            ]
        return self._request(
            23309,
            params,
            int
        )

    def delete_directory(
        self,
        dirname: str,
        time: datetime | None = None,
        device: Device | str = Device.INTERNAL
    ) -> GeoComResponse[int]:
        """
        RPC 23315, ``FTR_DeleteDir``

        .. versionadded:: GeoCom-VivaTPS

        Deletes one or more directories. Wildcards can be used to delete
        multiple items. If a date is given, only directories older than
        that date are deleted.

        Parameters
        ----------
        dirname : str
            Directory name.
        time : datetime | None, optional
            Deletion limit date, by default None
        device : Device | str, optional
            Memory device, by default PCPARD

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Number of directories deleted.
            Error codes:
                - ``IVPARAM``: Memory device unavailable, or cannot find
                  file path.

        See Also
        --------
        list

        """
        _device = toenum(Device, device)
        _filetype = File.DATABASE

        if time is None:
            params = [
                _device.value, _filetype.value,
                Byte(0), Byte(0), Byte(0),
                dirname
            ]
        else:
            params = [
                _device.value, _filetype.value,
                Byte(time.day), Byte(time.month), Byte(time.year - 2000),
                dirname
            ]
        return self._request(
            23315,
            params,
            int
        )

    def setup_large_download(
        self,
        filename: str,
        blocksize: int,
        device: Device | str = Device.INTERNAL,
        filetype: File | str = File.UNKNOWN
    ) -> GeoComResponse[int]:
        """
        RPC 23313, ``FTR_SetupDownloadLarge``

        .. versionadded:: GeoCom-VivaTPS

        Prepares download of the specified large file of the specified
        type, on the selected memory device.

        Parameters
        ----------
        filename : str
            File name (or full path if type is unknown).
        blocksize : int
            Download data block size.
        device : Device | str, optional
            Memory device, by default INTERNAL
        filetype : File | str, optional
            File type, by default UNKNOWN

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Number of download blocks needed.
            Error codes:
                - ``IVPARAM``: Memory device unavailable, or cannot find
                  file path.
                - ``NOTOK``: Setup already exists, previous setup was not
                  aborted.
                - ``FTR_INVALIDINPUT``: Block size too big.
                - ``FTR_FILEACCESS``: File access error.

        See Also
        --------
        download_large
        abort_download

        """
        _device = toenum(Device, device)
        _filetype = toenum(File, filetype)
        return self._request(
            23313,
            [_device.value, _filetype.value, filename, blocksize],
            int
        )

    def download_large(
        self,
        block: int
    ) -> GeoComResponse[str]:
        """
        RPC 23314, ``FTR_DownloadXL``

        .. versionadded:: GeoCom-VivaTPS

        Downloads a single data block of a previously defined large file
        download sequence.

        Parameters
        ----------
        block : int
            Number of data block to download.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Data block as hexadecimal byte string.
            Error codes:
                - ``FTR_MISSINGSETUP``: No active download setup.
                - ``FTR_INVALIDINPUT``: First block is missing.
                - ``FTR_FILEACCESS``: File access error.

        See Also
        --------
        setup_large_download
        abort_download

        """
        return self._request(
            23314,
            [block],
            parsestr
        )



"""
Description
===========

Module: ``geocompy.geo.mot``

Definitions for the GeoCom Motorization subsystem.

Types
-----

- ``GeoComMOT``

"""
from __future__ import annotations

from math import pi
from typing import SupportsFloat

from ..data import (
    toenum,
    enumparser
)
from .gcdata import (
    ATRLock,
    Controller,
    Stop
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComMOT(GeoComSubsystem):
    """
    Motorization subsystem of the GeoCom protocol.

    This subsystem provides access to motoriztaion parameters and control.

    """

    def get_lockon_status(self) -> GeoComResponse[ATRLock]:
        """
        RPC 6021, ``MOT_ReadLockStatus``

        Gets the current status of the ATR target lock.

        Returns
        -------
        GeoComResponse
            Params:
                - `ATRLock`: ATR lock status.
            Error codes:
                - ``NOT_IMPL``: Motorization not available.

        See Also
        --------
        aut.lock_in

        """
        return self._request(
            6021,
            parsers=enumparser(ATRLock)
        )

    def start_controller(
        self,
        mode: Controller | str = Controller.MANUAL
    ) -> GeoComResponse[None]:
        """
        RPC 6001, ``MOT_StartController``

        Starts the motor controller in the specified mode.

        Parameters
        ----------
        mode : Controller | str, optional
            Controller mode, by default Controller.MANUAL

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Control mode is not appropriate for velocity
                  control.
                - ``NOT_IMPL``: Motorization not available.
                - ``MOT_BUSY``: Subsystem is busy, controller already
                  started.
                - ``MOT_UNREADY``: Subsystem is not initialized.

        See Also
        --------
        set_velocity
        stop_controller

        """
        _mode = toenum(Controller, mode)
        return self._request(
            6001,
            [_mode.value]
        )

    def stop_controller(
        self,
        mode: Stop | str = Stop.NORMAL
    ) -> GeoComResponse[None]:
        """
        RPC 6002, ``MOT_StopController``

        Stops the active motor controller mode.

        Parameters
        ----------
        mode : Stop | str, optional
            Controller mode, by default Stop.NORMAL

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``MOT_NOT_BUSY``: Controller is not active.

        See Also
        --------
        set_velocity
        start_controller
        aus.switch_user_lock

        """
        _mode = toenum(Stop, mode)
        return self._request(
            6002,
            [_mode.value]
        )

    def set_velocity(
        self,
        hz: SupportsFloat,
        v: SupportsFloat
    ) -> GeoComResponse[None]:
        """
        RPC 6004, ``MOT_SetVelocity``

        Starts the motors at a constant speed. The motor controller must
        be set accordingly in advance.

        Parameters
        ----------
        hz : SupportsFloat
            Horizontal angle to turn in a second [-pi; +pi]rad.
        v : SupportsFloat
            Vertical angle to turn in a second [-pi; +pi]rad.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Velocities not within acceptable range.
                - ``MOT_NOT_CONFIG``: Motor controller was not started,
                  or is already busy with continuous task.
                - ``MOT_NOT_OCOST``: Controller is not set to constant
                  speed.
                - ``NOT_IMPL``: Motorization is not available.

        Note
        ----
        Instruments with piezo motors support velocities up to 180 deg/sec.
        Stations with traditional motor drives support velocoties up to
        45 deg/sec.

        See Also
        --------
        set_velocity
        start_controller
        aus.switch_user_lock

        """
        _horizontal = min(pi, max(-pi, float(hz)))
        _vertical = min(pi, max(-pi, float(v)))
        return self._request(
            6004,
            [_horizontal, _vertical]
        )



"""
Description
===========

Module: ``geocompy.geo.aut``

Definitions for the GeoCom Automation subsystem.

Types
-----

- ``GeoComAUT``

"""
from __future__ import annotations

from typing import SupportsFloat

from ..data import (
    Angle,
    toenum,
    enumparser,
    parsebool
)
from .gcdata import (
    Adjust,
    ATR,
    Position,
    Turn,
    Camera
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComAUT(GeoComSubsystem):
    """
    Automation subsystem of the GeoCom protocol.

    This subsystem controls most of the motorized functions of
    a total station, such as positioning of the axes, target search,
    target lock, etc.

    """

    def get_atr_status(self) -> GeoComResponse[bool]:
        """
        RPC 9019, ``AUT_GetATRStatus``

        .. deprecated:: GeoCom-TPS1100-1.04
            The command is still available, but should not be used with
            instruments that support the new `aus.get_user_atr_state`
            command.

        .. versionremoved:: GeoCom-TPS1200


        Gets whether or not the ATR mode is active.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: ATR mode is active.

            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        This command does not indicate if the ATR has acquired a prism.

        See Also
        --------
        switch_atr
        """
        return self._request(
            9019,
            parsers=parsebool
        )

    def switch_atr(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 9018, ``AUT_SetATRStatus``

        .. deprecated:: GeoCom-TPS1100-1.04
            The command is still available, but should not be used with
            instruments that support the new `aus.switch_user_atr`
            command.

        .. versionremoved:: GeoCom-TPS1200

        Activates or deactivates the ATR mode.

        Parameters
        ----------
        activate : bool
            Set ATR to active.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        If LOCK mode is active when the ATR is activated, then LOCK mode
        changes to ATR mode.

        If the ATR is deactivated, the LOCK mode does not change.

        See Also
        --------
        get_atr_status
        get_lock_status
        switch_lock
        """
        return self._request(9018, [activate])

    def get_lock_status(self) -> GeoComResponse[bool]:
        """
        RPC 9021, ``AUT_GetLockStatus``

        .. deprecated:: GeoCom-TPS1100-1.04
            The command is still available, but should not be used with
            instruments that support the new `aus.get_user_lock_state`
            command.

        .. versionremoved:: GeoCom-TPS1200

        Gets whether or not the lock mode is active.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Lock mode is active.

            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        See Also
        --------
        switch_lock
        mot.get_lockon_status
        """
        return self._request(
            9021,
            parsers=parsebool
        )

    def switch_lock(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 9020, ``AUT_SetLockStatus``

        .. deprecated:: GeoCom-TPS1100-1.04
            The command is still available, but should not be used with
            instruments that support the new `aus.switch_user_lock`
            command.

        .. versionremoved:: GeoCom-TPS1200

        Activates or deactivates the LOCK mode.

        Parameters
        ----------
        activate : bool
            LOCK state to set

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        Activating the LOCK mode does not mean that the instrument is
        automatically locked onto a prism.

        See Also
        --------
        get_lock_status
        get_atr_status
        lock_in
        """
        return self._request(
            9020,
            [activate]
        )

    def get_tolerance(self) -> GeoComResponse[tuple[Angle, Angle]]:
        """
        RPC 9008, ``AUT_ReadTol``

        Gets the positioning tolerances on the Hz and V axes.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

            Params:
                - `Angle`: Horizontal tolerance.
                - `Angle`: Vertical tolerance.

        See Also
        --------
        set_tolerance
        """
        return self._request(
            9008,
            parsers=(Angle.parse, Angle.parse)
        )

    def set_tolerance(
        self,
        hz: SupportsFloat,
        v: SupportsFloat
    ) -> GeoComResponse[None]:
        """
        RPC 9007, ``AUT_SetTol``

        Sets the positioning tolerances on the Hz and V axes.

        Parameters
        ----------
        hz : SupportsFloat
            Horizontal tolerance.
        v : SupportsFloat
            Vertical tolerance.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``IVPARAM``: Tolerances are out of the valid range.
                - ``MOT_UNREADY``: Instrument has no motorization.

        See Also
        --------
        get_tolerance
        """
        return self._request(9007, [float(hz), float(v)])

    def get_timeout(self) -> GeoComResponse[tuple[float, float]]:
        """
        RPC 9012, ``AUT_ReadTimeout``

        Gets the positioning timeout for the Hz and V axes.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

            Params:
                - `float`: Horizontal timeout [sec].
                - `float`: Vertical timeout [sec].

        See Also
        --------
        set_timeout
        """
        return self._request(
            9012,
            parsers=(float, float)
        )

    def set_timeout(
        self,
        hz: float,
        v: float
    ) -> GeoComResponse[None]:
        """
        RPC 9011, ``AUT_SetTimeout``

        Sets the positioning timeout for the Hz and V axes.

        Parameters
        ----------
        hz : float
            Horizontal timeout [sec].
        v : float
            Vertical timeout [sec]

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``IVPARAM``: Timeout values are not in the [7; 60] range.

        See Also
        --------
        get_timeout
        """
        return self._request(
            9011,
            [hz, v]
        )

    def turn_to(
        self,
        hz: SupportsFloat,
        v: SupportsFloat,
        posmode: Position | str = Position.NORMAL,
        atrmode: ATR | str = ATR.POSITION
    ) -> GeoComResponse[None]:
        """
        RPC 9027, ``AUT_MakePositioning``

        Turns the telescope to the specified angular positions.

        Parameters
        ----------
        hz : SupportsFloat
            Horizontal position.
        v : SupportsFloat
            Vertical position.
        posmode : Position | str, optional
            Positioning precision mode, by default Position.NORMAL
        atrmode : ATR | str, optional
            ATR mode, by default ATR.POSITION

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid parameter
                - ``AUT_TIMEOUT``: Positioning timed out.
                - ``AUT_MOTOR_ERROR``: Instrument has no motorization.
                - ``AUT_ANGLE_ERROR``: Angle measurement error.
                - ``AUT_INACC``: Inexact position.
                - ``ABORT``: Function aborted.
                - ``COM_TIMEDOUT``: Communication timeout.
                - ``AUT_NO_TARGET``: No ATR target found.
                - ``AUT_MULTIPLE_TARGETS``: Multiple ATR targets found.
                - ``AUT_BAD_ENVIRONMENT``: Inadequate environmental
                  conditions.
                - ``AUT_ACCURACY``: Position is not within tolerances.
                  Repeat positioning!
                - ``AUT_DEV_ERROR``: Angle deviation calculation error.
                  Repeat positioning!
                - ``AUT_NOT_ENABLED``: ATR mode is not active.

        See Also
        --------
        get_atr_status
        switch_atr
        get_lock_status
        switch_lock
        get_tolerance
        set_tolerance
        get_timeout
        set_timeout
        com.get_timeout
        com.set_timeout
        """
        _posmode = toenum(Position, posmode)
        _atrmode = toenum(ATR, atrmode)
        return self._request(
            9027,
            [float(hz), float(v), _posmode.value, _atrmode.value, 0]
        )

    def change_face(
        self,
        posmode: Position | str = Position.NORMAL,
        atrmode: ATR | str = ATR.POSITION
    ) -> GeoComResponse[None]:
        """
        RPC 9028, ``AUT_ChangeFace``

        Turns the telescope to the opposite face.

        Parameters
        ----------
        posmode : Position | str, optional
            Positioning precision mode, by default Position.NORMAL
        atrmode : ATR | str, optional
            ATR mode, by default ATR.POSITION

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid parameter
                - ``AUT_TIMEOUT``: Positioning timed out.
                - ``AUT_MOTOR_ERROR``: Instrument has no motorization.
                - ``AUT_ANGLE_ERROR``: Angle measurement error.
                - ``AUT_INACC``: Inexact position.
                - ``ABORT``: Function aborted.
                - ``COM_TIMEDOUT``: Communication timeout.
                - ``AUT_NO_TARGET``: No ATR target found.
                - ``AUT_MULTIPLE_TARGETS``: Multiple ATR targets found.
                - ``AUT_BAD_ENVIRONMENT``: Inadequate environmental
                  conditions.
                - ``AUT_ACCURACY``: Position is not within tolerances.
                  Repeat positioning!
                - ``AUT_DEV_ERROR``: Angle deviation calculation error.
                  Repeat positioning!
                - ``AUT_NOT_ENABLED``: ATR mode is not active.

        See Also
        --------
        get_atr_status
        switch_atr
        get_lock_status
        switch_lock
        get_tolerance
        set_tolerance
        get_timeout
        set_timeout
        com.get_timeout
        com.set_timeout
        tmc.get_face
        """
        _posmode = toenum(Position, posmode)
        _atrmode = toenum(ATR, atrmode)
        return self._request(
            9028,
            [_posmode.value, _atrmode.value, 0]
        )

    def fine_adjust(
        self,
        width: SupportsFloat,
        height: SupportsFloat
    ) -> GeoComResponse[None]:
        """
        RPC 9037, ``AUT_FineAdjust``

        Precisely targets a prism. If the prism is not within the view of
        the ATR, a target search is executed in the specified window.

        Parameters
        ----------
        width : SupportsFloat
            Width of target search window.
        height : SupportsFloat
            Heigth of target search window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``IVPARAM``: Invalid parameter
                - ``AUT_TIMEOUT``: Positioning timed out.
                - ``AUT_MOTOR_ERROR``: Instrument has no motorization.
                - ``FATAL``: Fatal error.
                - ``ABORT``: Function aborted.
                - ``COM_TIMEDOUT``: Communication timeout.
                - ``AUT_NO_TARGET``: No ATR target found.
                - ``AUT_MULTIPLE_TARGETS``: Multiple ATR targets found.
                - ``AUT_BAD_ENVIRONMENT``: Inadequate environmental
                  conditions.
                - ``AUT_DEV_ERROR``: Angle deviation calculation error.
                  Repeat positioning!
                - ``AUT_NOT_ENABLED``: ATR mode is not active.
                - ``AUT_DETECTOR_ERROR``: Error in target acquisition.

        See Also
        --------
        get_atr_status
        switch_atr
        get_fine_adjust_mode
        set_fine_adjust_mode
        """
        return self._request(
            9037,
            [float(width), float(height), 0]
        )

    def search(
        self,
        width: SupportsFloat,
        height: SupportsFloat
    ) -> GeoComResponse[None]:
        """
        RPC 9029, ``AUT_Search``

        Search for target in the specified search window. The search is
        terminated once a prism appears in the view of the ATR. Fine
        adjustment must be executed afterwards.

        Parameters
        ----------
        width : SupportsFloat
            Width of target search window.
        height : SupportsFloat
            Heigth of target search window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid parameter
                - ``AUT_MOTOR_ERROR``: Instrument has no motorization.
                - ``TMC_NO_FULL_CORRECTION``: Instrument might not be
                  properly levelled.
                - ``FATAL``: Fatal error.
                - ``ABORT``: Function aborted.
                - ``COM_TIMEDOUT``: Communication timeout.
                - ``AUT_NO_TARGET``: No ATR target found.
                - ``AUT_BAD_ENVIRONMENT``: Inadequate environmental
                  conditions.
                - ``AUT_NOT_ENABLED``: ATR mode is not active.
                - ``AUT_DETECTOR_ERROR``: Error in target acquisition.

        See Also
        --------
        get_atr_status
        switch_atr
        fine_adjust
        """
        return self._request(
            9029,
            [float(width), float(height), 0]
        )

    def get_fine_adjust_mode(self) -> GeoComResponse[Adjust]:
        """
        RPC 9030, ``AUT_GetFineAdjustMode``

        Gets the fine adjustment mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `Adjust`: Current fine adjustment mode.

        See Also
        --------
        set_fine_adjust_mode
        """
        return self._request(
            9030,
            parsers=enumparser(Adjust)
        )

    def set_fine_adjust_mode(
        self,
        mode: Adjust | str
    ) -> GeoComResponse[None]:
        """
        RPC 9031, ``AUT_SetFineAdjustMode``

        Sets the fine adjustment mode.

        Parameters
        ----------
        mode : Adjust | str

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``IVPARAM``: Invalid mode

        See Also
        --------
        get_fine_adjust_mode
        """
        _mode = toenum(Adjust, mode)
        return self._request(
            9031,
            [_mode.value]
        )

    def lock_in(self) -> GeoComResponse[None]:
        """
        RPC 9013, ``AUT_LockIn``

        Locks onto target prism and starts tracking. LOCK mode must be
        active, and fine adjustment must have been completed, before
        executing this command.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``AUT_NOT_ENABLED``: ATR mode is not active.
                - ``AUT_MOTOR_ERROR``: Instrument has not motorization.
                - ``AUT_DETECTOR_ERROR``: Error in target acquisition.
                - ``AUT_NO_TARGET``: No ATR target found.
                - ``AUT_BAD_ENVIRONMENT``: Inadequate environmental
                  conditions.

        See Also
        --------
        get_lock_status
        switch_lock
        mot.get_lockon_status
        """
        return self._request(9013)

    def get_search_area(
        self
    ) -> GeoComResponse[tuple[Angle, Angle, Angle, Angle, bool]]:
        """
        RPC 9042, ``AUT_GetSearchArea``

        .. versionadded:: GeoCom-TPS1100-1.04

        Gets current position and size of the PowerSearch window.

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal center of window.
                - `Angle`: Vertical center of window.
                - `Angle`: Width of window.
                - `Angle`: Height of window.
                - `bool`: If window is enabled.
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

        See Also
        --------
        set_search_area
        bap.search_target
        """
        return self._request(
            9042,
            parsers=(
                Angle.parse,
                Angle.parse,
                Angle.parse,
                Angle.parse,
                parsebool
            )
        )

    def set_search_area(
        self,
        hz: SupportsFloat,
        v: SupportsFloat,
        width: SupportsFloat,
        height: SupportsFloat,
        enabled: bool = True
    ) -> GeoComResponse[None]:
        """
        RPC 9043, ``AUT_SetSearchArea``

        .. versionadded:: GeoCom-TPS1100-1.04

        Sets position and size of the PowerSearch window.

        Parameters
        ----------
        hz : SupportsFloat
            Horizontal center of search window.
        v : SupportsFloat
            Vertical center of search window.
        width : SupportsFloat
            Width of search window.
        height : SupportsFloat
            Height of search window.
        enabled : bool
            Activation state of search window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

        See Also
        --------
        get_search_area
        bap.search_target
        """
        return self._request(
            9043,
            [float(hz), float(v), float(width), float(height), float(enabled)]
        )

    def get_spiral(self) -> GeoComResponse[tuple[Angle, Angle]]:
        """
        RPC 9040, ``AUT_GetUserSpiral``

        .. versionadded:: GeoCom-TPS1100-1.04

        Gets the size of the PowerSearch window.

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Width of window.
                - `Angle`: Height of window.
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

        See Also
        --------
        set_spiral
        bap.search_target
        """
        return self._request(
            9040,
            parsers=(Angle.parse, Angle.parse)
        )

    def set_spiral(
        self,
        width: SupportsFloat,
        height: SupportsFloat
    ) -> GeoComResponse[None]:
        """
        RPC 9041, ``AUT_SetUserSpiral``

        .. versionadded:: GeoCom-TPS1100-1.04

        Sets the size of the PowerSearch window.

        Parameters
        ----------
        width : SupportsFloat
            Width of the search window.
        height : SupportsFloat
            Height of the search window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

        See Also
        --------
        get_spiral
        bap.search_target
        """
        return self._request(
            9041,
            [float(width), float(height)]
        )

    def switch_powersearch_range(
        self,
        enable: bool
    ) -> GeoComResponse[None]:
        """
        RPC 9048, ``AUT_PS_EnableRange``

        Enables or disables the PowerSearch window and range limit.

        Parameters
        ----------
        enable : bool
            Enable predefined PowerSearch window and range limits.
            Defaults to [0; 400] range when disabled.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.

        See Also
        --------
        set_powersearch_range
        set_search_area
        """
        return self._request(
            9048,
            [enable]
        )

    def set_powersearch_range(
        self,
        closest: int,
        farthest: int
    ) -> GeoComResponse[None]:
        """
        RPC 9044, ``AUT_PS_SetRange``

        .. versionadded:: GeoCom-TPS1200

        Sets the PowerSearch range limits.

        Parameters
        ----------
        closest : int
            Minimum distance to prism [0; ...].
        farthest : int
            Maxmimum distance to prism [closest + 10; 400].

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``IVPARAM``: Invalid parameters.

        See Also
        --------
        switch_powersearch_range
        powersearch_window
        set_search_area
        """
        return self._request(
            9047,
            [closest, farthest]
        )

    def powersearch_window(self) -> GeoComResponse[None]:
        """
        RPC 9052, ``AUT_PS_SearchWindow``

        .. versionadded:: GeoCom-TPS1200

        Executes PowerSearch in the predefined search window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``AUT_NO_WORKING_AREA``: Search window is not defined.
                - ``AUT_NO_TARGET``: Target was not found.

        See Also
        --------
        switch_powersearch_range
        set_powersearch_range
        powersearch_next
        set_search_area
        """
        return self._request(9052)

    def powersearch_next(
        self,
        direction: Turn | str,
        swing: bool
    ) -> GeoComResponse[None]:
        """
        RPC 9051, ``AUT_PS_SearchNext``

        .. versionadded:: GeoCom-TPS1200

        Executes 360° default PowerSearch to find the next target.

        Parameters
        ----------
        direction : Turn | str
            Turning direction during PowerSearch.
        swing : bool
            Search starts -10 GON to the given turn direction.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: GeoCom Robotic license not found.
                - ``AUT_NO_TARGET``: Target was not found.
                - ``IVPARAM``: Invalid parameters.

        See Also
        --------
        switch_powersearch_range
        powersearch_window
        """
        _direction = toenum(Turn, direction)
        return self._request(
            9051,
            [_direction.value, swing]
        )

    def switch_lock_onthefly(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 9103, ``AUT_SetLockFlyMode``

        .. versionadded:: GeoCom-VivaTPS

        Sets the state of on-the-fly mode for the lock mode.

        Parameters
        ----------
        enabled : bool
            Enable on-the-fly lock mode.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_lock_onthefly_status
        """
        return self._request(
            9103,
            [enabled]
        )

    def get_lock_onthefly_status(self) -> GeoComResponse[bool]:
        """
        RPC 9102, ``AUT_GetLockFlyMode``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current state of the on-the-fly lock mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: On-the-fly lock mode is enabled.

        See Also
        --------
        get_lock_onthefly_status
        """
        return self._request(
            9102,
            parsers=parsebool
        )

    def aim_at_pixel(
        self,
        x: int,
        y: int,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 9081, ``AUT_CAM_PositToPixelCoord``

        .. versionadded:: GeoCom-VivaTPS

        Turns the instrument to face the coordinates specified in the
        image coordinates.

        Parameters
        ----------
        x : int
            Horizontal pixel coordinate.
        y : int
            Vertical pixel coordinate.
        camera : Camera, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.
                - ``AUT_SIDECOVER_ERR``: Sidecover is open.

        """
        _camera = toenum(Camera, camera)
        return self._request(
            9081,
            [_camera.value, x, y]
        )



import re

from geocompy.communication import Connection
from geocompy.data import gsiword
from geocompy.gsi.gsitypes import GsiOnlineType

from helpers import faulty_parser


def gsiparser(value: str) -> int:
    value = value.strip("* ")
    return int(value[6:])


class DummyGsiOnlineConnection(Connection):
    _CONF = re.compile(r"^CONF/\d+$")
    _SET = re.compile(r"^SET/\d+/\d+$")
    _GET = re.compile(r"^GET/[M,I,C]/WI\d+$")
    _PUT = re.compile(
        r"^PUT/"
        r"\*?"
        r"(?:[0-9\.]{6})"
        r"(?:\+|\-)"
        r"(?:[a-zA-Z0-9]{8}|[a-zA-Z0-9]{16}) $"
    )

    def __init__(self, gsi16: bool = False):
        self._gsi16 = gsi16

    def send(self, value: str) -> None:
        pass

    def receive(self) -> str:
        return ""

    def exchange(self, cmd: str) -> str:
        if self._CONF.match(cmd) and cmd != "CONF/0":
            if cmd == "CONF/137":
                return f"0137/{self._gsi16:04d}"
            return f"{cmd.split('/')[1].zfill(4)}/0000"
        elif self._SET.match(cmd) and cmd != "SET/0/0":
            return "?"
        elif self._GET.match(cmd) and cmd != "GET/I/WI0":
            wi = int(cmd.split("/")[-1].removeprefix("WI"))
            word = gsiword(
                wi,
                "1",
                gsi16=self._gsi16
            )

            return word
        elif self._PUT.match(cmd) and cmd != "PUT/0.....+00000000 ":
            return "?"
        elif cmd in ("a", "b", "c", "BEEP/0", "BEEP/1", "BEEP/2"):
            return "?"

        return "@W427"

    def reset(self) -> None:
        pass

    def close(self) -> None:
        pass

    def is_open(self) -> bool:
        return True


class GsiOnlineTester:
    @staticmethod
    def test_request(instrument: GsiOnlineType) -> None:
        response = instrument.request("d")
        assert not response.value
        response = instrument.request("a")
        assert response.value

    @staticmethod
    def test_setrequest(instrument: GsiOnlineType) -> None:
        response = instrument.setrequest(0, 0)
        assert not response.value
        assert response.comment == "INSTRUMENT"
        assert response.response == "@W427"

        response = instrument.setrequest(1, 1)
        assert response.value

    @staticmethod
    def test_confrequest(instrument: GsiOnlineType) -> None:
        response = instrument.confrequest(0, int)
        assert not response.value
        assert response.comment == "INSTRUMENT"
        assert response.response == "@W427"

        response = instrument.confrequest(1, faulty_parser)
        assert response.value is None
        assert response.comment == "PARSE"

        response = instrument.confrequest(1, int)
        assert response.comment == ""
        assert response.response == "0001/0000"
        assert response.value == 0

    @staticmethod
    def test_putrequest(instrument: GsiOnlineType) -> None:
        response = instrument.putrequest(0, "0.....+00000000 ")
        assert not response.value
        assert response.comment == "INSTRUMENT"
        assert response.response == "@W427"

        response = instrument.putrequest(1, "1.....+00000001 ")
        assert response.value

    @staticmethod
    def test_getrequest(instrument: GsiOnlineType) -> None:
        response = instrument.getrequest("I", 0, int)
        assert response.value is None
        assert response.response == "@W427"

        response = instrument.getrequest("I", 1, faulty_parser)
        assert response.value is None
        assert response.comment == "PARSE"

        response = instrument.getrequest("I", 1, gsiparser)
        assert response.value == 1



"""
Description
===========

Module: ``geocompy.gis.gsitypes``

The GSI Online types module provides type definitions and general
constants, that are relevant to the GSI Online protocol.

Types
-----

- ``GsiOnlineResponse``
- ``GsiOnlineType``
- ``GsiOnlineSubsystem``
"""
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Callable, Literal


_T = TypeVar("_T")


class GsiOnlineResponse(Generic[_T]):
    """Container class for parsed GSI Online responses."""

    def __init__(
        self,
        desc: str,
        cmd: str,
        response: str,
        value: _T | None,
        comment: str = ""
    ):
        """
        Parameters
        ----------
        desc : str
            Description of the GSI Online command, that invoked this
            response.
        cmd : str
            Full, serialized command, that invoked this response.
        response : str
            Full, received response.
        value
            Parsed response value. The content is dependent on the
            executed command.
        comment : str
            Additional comment (e.g. explanation of an error).
        """
        self.desc: str = desc
        """Description of the GSI Online command, that invoked this
        response."""
        self.cmd: str = cmd
        """Full, serialized command, that invoked this response."""
        self.response: str = response
        """Full, received response."""
        self.value: _T | None = value
        """Parsed response value. The content is dependent on the
        executed command."""
        self.comment: str = comment
        """Additional comment (e.g. explanation of an error)."""

    def __str__(self) -> str:
        success = (
            "success"
            if self.value is not None
            else f"fail ({self.comment})"
        )
        return (
            f"GsiOnlineResponse({self.desc}) "
            f"{success}, "
            f"value: {self.value}, "
            f"(cmd: '{self.cmd}', response: '{self.response}')"
        )

    def __bool__(self) -> bool:
        return self.value is not None


class GsiOnlineType(ABC):
    """
    Interface definition for the GSI Online protocol handler type.
    """
    @property
    @abstractmethod
    def is_client_gsi16(self) -> bool: ...

    @is_client_gsi16.setter
    @abstractmethod
    def is_client_gsi16(self, value: bool) -> None: ...

    @abstractmethod
    def setrequest(
        self,
        param: int,
        value: int
    ) -> GsiOnlineResponse[bool]: ...

    @abstractmethod
    def confrequest(
        self,
        param: int,
        parser: Callable[[str], _T]
    ) -> GsiOnlineResponse[_T | None]: ...

    @abstractmethod
    def putrequest(
        self,
        wordindex: int,
        word: str
    ) -> GsiOnlineResponse[bool]: ...

    @abstractmethod
    def getrequest(
        self,
        mode: Literal['I', 'M', 'C'],
        wordindex: int,
        parser: Callable[[str], _T]
    ) -> GsiOnlineResponse[_T | None]: ...

    @abstractmethod
    def request(
        self,
        cmd: str,
        desc: str = ""
    ) -> GsiOnlineResponse[bool]: ...


class GsiOnlineSubsystem:
    """
    Base class for GSI Online subsystems.
    """

    def __init__(self, parent: GsiOnlineType):
        """
        Parameters
        ----------
        parent : GsiOnlineType
            The parent protocol instance of this subsystem.
        """
        self._parent: GsiOnlineType = parent
        """Parent protocol instance"""
        self._setrequest = self._parent.setrequest
        """Shortcut to the `setrequest` method of the parent protocol."""
        self._confrequest = self._parent.confrequest
        """Shortcut to the `confrequest` method of the parent protocol."""
        self._putrequest = self._parent.putrequest
        """Shortcut to the `putrequest` method of the parent protocol."""
        self._getrequest = self._parent.getrequest
        """Shortcut to the `getrequest` method of the parent protocol."""


param_descriptions: dict[int, str] = {
    30: "Beep intensity",
    31: "Display illumination",
    32: "Display constrast",
    41: "Distance unit",
    42: "Temperature unit",
    51: "Decimals displayed",
    70: "Serial speed",
    71: "Parity",
    73: "Terminator",
    75: "Protocol",
    76: "Recording device",
    78: "Send delay",
    90: "Battery level",
    91: "Internal temperature",
    95: "Auto off",
    106: "Display heater",
    125: "Earth curvature correction",
    127: "Staff direction",
    137: "GSI type",
    138: "Code recording mode"
}
"""Mapping of parameter indices to short descriptions."""


word_descriptions: dict[int, str] = {
    11: "Point ID",
    71: "Note",
    560: "Time",
    561: "Date",
    562: "Year",
    32: "Distance",
    330: "Reading",
    95: "Internal temperature",
    12: "Serial number",
    13: "Instrument type",
    17: "Full date",
    19: "Date and time",
    599: "Software version"
}
"""Mapping of GSI word indices to short descriptions."""



"""
Description
===========

Module: ``geocompy.geo.wir``

Definitions for the GeoCom Word Index registration subsystem.

Types
-----

- ``GeoComWIR``

"""
from __future__ import annotations

from ..data import (
    enumparser,
    toenum
)
from .gcdata import Format
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComWIR(GeoComSubsystem):
    """
    Word Index registration subsystem of the GeoCom protocol.
    This subsystem is responsible for the GSI data recording operations.

    .. versionremoved:: GeoCom-TPS1200
    """

    def get_recording_format(self) -> GeoComResponse[Format]:
        """
        RPC 8011, ``WIR_GetRecFormat``

        Retrieves the current data recording format.

        Returns
        -------
        GeoComResponse
            Params:
                - `Format`: GSI version used in data recording.

        """
        return self._request(
            8011,
            parsers=enumparser(Format)
        )

    def set_recording_format(
        self,
        format: Format | str
    ) -> GeoComResponse[None]:
        """
        RPC 8012, ``WIR_SetRecFormat``

        Sets the data recording format.

        Parameters
        ----------
        format : Format | str
            GSI format to use in data recording.

        Returns
        -------
        GeoComResponse

        """
        _format = toenum(Format, format)
        return self._request(
            8012,
            [_format.value]
        )



"""
Description
===========

Module: ``geocompy.geo.csv``

Definitions for the GeoCom Central services subsystem.

Types
-----

- ``GeoComCSV``

"""
from __future__ import annotations

from datetime import datetime

from ..data import (
    Byte,
    parsestr,
    parsebool,
    enumparser,
    toenum
)
from .gcdata import (
    Capabilities,
    DeviceClass,
    PowerSource,
    Reflectorless,
    Property
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComCSV(GeoComSubsystem):
    """
    Central services subsystem of the GeoCom protocol.

    This subsystem contains functions to maintain centralised data
    and configuration of the instruments.

    """

    def get_serial_number(self) -> GeoComResponse[int]:
        """
        RPC 5003, ``CSV_GetInstrumentNo``

        Gets the serial number of the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Serial number.

        """
        return self._request(
            5003,
            parsers=int
        )

    def get_instrument_name(self) -> GeoComResponse[str]:
        """
        RPC 5004, ``CSV_GetInstrumentName``

        Gets the name of the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Instrument name.

        """
        return self._request(
            5004,
            parsers=parsestr
        )

    def get_user_instrument_name(self) -> GeoComResponse[str]:
        """
        RPC 5006, ``CSV_GetUserInstrumentName``

        .. versionremoved:: GeoCom-TPS1100

        Gets the user defined name of the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Instrument name.

        """
        return self._request(
            5006,
            parsers=parsestr
        )

    def set_user_instrument_name(
        self,
        name: str
    ) -> GeoComResponse[None]:
        """
        RPC 5005, ``CSV_SetUserInstrumentName``

        .. versionremoved:: GeoCom-TPS1100

        Parameters
        ----------
        name : str
            Instrument name.

        Returns
        -------
        GeoComResponse
        """
        return self._request(
            5005,
            [name]
        )

    def get_instrument_configuration(
        self
    ) -> GeoComResponse[tuple[DeviceClass, Capabilities]]:
        """
        RPC 5035, ``CSV_GetDeviceConfig``

        Gets class of the instrument, as well as information about
        the capatilities of the configuration.

        Returns
        -------
        GeoComResponse
            Params:
                - `DeviceClass`: Class of the instrument.
                - `Capabilities`: Configuration of the components.
            Error codes:
                - ``UNDEFINED``: Precision class is undefined.

        """
        return self._request(
            5035,
            parsers=(
                enumparser(DeviceClass),
                enumparser(Capabilities)
            )
        )

    def get_datetime(self) -> GeoComResponse[datetime]:
        """
        RPC 5008, ``CSV_GetDateTime``

        Gets the current date and time set on the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `datetime`: Current date and time.

        See Also
        --------
        set_datetime

        """
        def make_datetime(
            params: tuple[int, Byte, Byte, Byte, Byte, Byte] | None
        ) -> datetime | None:
            if params is None:
                return None

            return datetime(
                params[0],
                int(params[1]),
                int(params[2]),
                int(params[3]),
                int(params[4]),
                int(params[5])
            )

        response: GeoComResponse[
            tuple[int, Byte, Byte, Byte, Byte, Byte]
        ] = self._request(
            5008,
            parsers=(
                int,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse
            )
        )

        return response.map_params(make_datetime)

    def set_datetime(
        self,
        time: datetime
    ) -> GeoComResponse[None]:
        """
        RPC 5007, ``CSV_SetDateTime``

        Sets the date and time on the instrument.

        Parameters
        ----------
        time : datetime
            New date and time to set.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_datetime

        """
        return self._request(
            5007,
            [
                time.year, Byte(time.month), Byte(time.day),
                Byte(time.hour), Byte(time.minute), Byte(time.second)
            ]
        )

    def get_firmware_version(self) -> GeoComResponse[tuple[int, int, int]]:
        """
        RPC 5034, ``CSV_GetSWVersion``

        Gets the system software version running on the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Release number.
                - `int`: Version number.
                - `int`: Subversion number.

        """
        return self._request(
            5034,
            parsers=(int, int, int)
        )

    def get_firmware_creation_date(self) -> GeoComResponse[str]:
        """
        RPC 5038, ``CSV_GetSWCreationDate``

        .. versionadded:: GeoComp-TPS1200

        Gets the creation date of the system software version running on
        the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Creation date.

        """
        # def transformer(value: str | None) -> datetime | None:
        #     if value is None:
        #         return None

        #     return datetime.strptime(value, "%Y-%m-%d")

        response = self._request(
            5038,
            parsers=str
        )
        return response

    def get_voltage_battery(self) -> GeoComResponse[float]:
        """
        RPC 5009, ``CSV_GetVBat``

        .. deprecated:: GeoCom-TPS1100-1.05
            The command is still available, but should not be used with
            instruments that support the new `check_power` command.

        .. versionremoved:: GeoCom-TPS1200

        Gets the voltage of the power supply.

        | 12,7 V < voltage            full
        | 12,4 V < voltage < 12,7 V   near full
        | 11,1 V < voltage < 12,4 V   good
        | 10,5 V < voltage < 11,1 V   empty
        |          voltage < 10,5 V   powered off

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Power source voltage [V].

        """
        return self._request(
            5009,
            parsers=float
        )

    def get_voltage_memory(self) -> GeoComResponse[float]:
        """
        RPC 5009, ``CSV_GetVMem``

        .. versionremoved:: GeoCom-TPS1200

        Gets the voltage of the memory backup power supply.

        Voltage above 3.1 V is OK.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Power source voltage [V].

        """
        return self._request(
            5009,
            parsers=float
        )

    def get_internal_temperature(self) -> GeoComResponse[float]:
        """
        RPC 5011, ``CSV_GetIntTemp``

        Gets internal temperature of the instrument, measured on the
        main board.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Internal temperature [°C].

        """
        return self._request(
            5011,
            parsers=float
        )

    def check_power(
        self
    ) -> GeoComResponse[tuple[int, PowerSource, PowerSource]]:
        """
        RPC 5039, ``CSV_CheckPower``

        .. versionadded:: GeoCom-TPS1100-1.05

        Gets the remaining capacity of the active power source.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Remaining capacity [%].
                - `PowerSource`: Active power source.
                - `PowerSource`: Suggested power source.

        """
        return self._request(
            5039,
            parsers=(
                int,
                enumparser(PowerSource),
                enumparser(PowerSource)
            )
        )

    def get_reflectorless_class(self) -> GeoComResponse[Reflectorless]:
        """
        RPC 5100, ``CSV_GetReflectorlessClass``

        .. versionadded:: GeoCom-TPS1200

        Gets the class of the reflectorless EDM module, if the instrument
        is equipped with one.

        Returns
        -------
        GeoComResponse
            Params:
                - `Reflectorless`: Class of the reflectorless EDM module.

        """
        return self._request(
            5100,
            parsers=enumparser(Reflectorless)
        )

    def get_datetime_precise(self) -> GeoComResponse[datetime]:
        """
        RPC 5117, ``CSV_GetDateTimeCentiSec``

        .. versionadded:: GeoCom-TPS1200

        Gets the current date and time set on the instrument in
        centiseconds resolution.

        Returns
        -------
        GeoComResponse
            Params:
                - `datetime`: Current date and time.

        See Also
        --------
        get_datetime
        set_datetime

        """
        def make_datetime(
            params: tuple[int, int, int, int, int, int, int] | None
        ) -> datetime | None:
            if params is None:
                return None

            return datetime(
                params[0],
                int(params[1]),
                int(params[2]),
                int(params[3]),
                int(params[4]),
                int(params[5]),
                int(params[5]) * 10000
            )

        response = self._request(
            5117,
            parsers=(
                int,
                int,
                int,
                int,
                int,
                int,
                int
            )
        )
        return response.map_params(make_datetime)

    def switch_startup_message(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 5155, ``CSV_SetStartupMessageMode``

        .. versionadded:: GeoCom-VivaTPS

        Enables or disables the startup message mode on the instrument.

        Parameters
        ----------
        enabled : bool
            Startup message mode is enabled.

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            5155,
            [enabled]
        )

    def get_startup_message_status(self) -> GeoComResponse[bool]:
        """
        RPC 5156, ``CSV_GetStartupMessageMode``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current status of the startup message mode on the
        instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Startup message is enabled.

        """
        return self._request(
            5156,
            parsers=parsebool
        )

    def switch_laserlot(
        self,
        active: bool
    ) -> GeoComResponse[None]:
        """
        RPC 5043, ``CSV_SwitchLaserlot``

        .. versionadded:: GeoCom-VivaTPS

        Sets the state of the laserlot.

        Parameters
        ----------
        active : bool
            Activate laserlot.

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            5043,
            [active]
        )

    def get_laserlot_status(self) -> GeoComResponse[bool]:
        """
        RPC 5042, ``CSV_GetLaserlotStatus``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current state of the laserlot.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Laserlot is active.

        """
        return self._request(
            5042,
            parsers=parsebool
        )

    def set_laserlot_intensity(
        self,
        intensity: int
    ) -> GeoComResponse[None]:
        """
        RPC 5040, ``CSV_SetLaserlotIntens``

        .. versionadded:: GeoCom-VivaTPS

        Sets the intensity of the laserlot.

        Parameters
        ----------
        intensity : int
            New laserlot intensity to set.

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            5040,
            [intensity]
        )

    def get_laserlot_intensity(self) -> GeoComResponse[int]:
        """
        RPC 5041, ``CSV_GetLaserlotIntens``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current intensity of the laserlot.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Current laserlot intensity.

        """
        return self._request(
            5041,
            parsers=int
        )

    def check_property(
        self,
        property: Property | str
    ) -> GeoComResponse[bool]:
        """
        RPC 5039, ``CSV_CheckProperty``

        .. versionadded:: GeoCom-VivaTPS

        Checks if a specific license is available on the instrument.

        Parameters
        ----------
        property : Property | str
            License to check.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: License is available.

        """
        _prop = toenum(Property, property)
        return self._request(
            5139,
            [_prop.value],
            parsebool
        )

    def get_voltage(self) -> GeoComResponse[int]:
        """
        RPC 5165, ``CSV_GetVoltage``

        .. versionadded:: GeoCom-VivaTPS

        Gets the instrument voltage.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Instrument voltage [mV].

        """
        return self._request(
            5165,
            parsers=int
        )

    def switch_charging(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 5161, ``CSV_SetCharging``

        .. versionadded:: GeoCom-VivaTPS

        Sets the state of the charger.

        Parameters
        ----------
        activate : bool
            Activate charger.

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            5161,
            [activate]
        )

    def get_charging_status(self) -> GeoComResponse[bool]:
        """
        RPC 5162, ``CSV_GetCharging``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current state of the charger.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Charger is active.

        """
        return self._request(
            5162,
            parsers=parsebool
        )

    def set_preferred_powersource(
        self,
        source: PowerSource | str
    ) -> GeoComResponse[None]:
        """
        RPC 5163, ``CSV_SetPreferredPowersource``

        .. versionadded:: GeoCom-VivaTPS

        Sets the preferred power source.

        Parameters
        ----------
        source : PowerSource | str
            New preferred power source to set.

        """
        _source = toenum(PowerSource, source)
        return self._request(
            5163,
            [_source.value]
        )

    def get_preferred_powersource(self) -> GeoComResponse[PowerSource]:
        """
        RPC 5164, ``CSV_GetPreferredPowersource``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current preferred power source.

        Returns
        -------
        GeoComResponse
            Params:
                - `PowerSource`: Preferred power source.

        """
        return self._request(
            5164,  # Mistyped as 5163 in the GeoCom reference
            parsers=enumparser(PowerSource)
        )

    def get_datetime_new(self) -> GeoComResponse[datetime]:
        """
        RPC 5051, ``CSV_GetDateTime2``

        .. versionadded:: GeoComp-TPS1200

        Gets the current date and time set on the instrument.

        Returns
        -------
        GeoComResponse
            Params:
                - `datetime`: Current date and time.

        See Also
        --------
        set_datetime_new

        """
        def make_datetime(
            params: tuple[int, int, int, int, int, int] | None
        ) -> datetime | None:
            if params is None:
                return None

            return datetime(
                params[0],
                params[1],
                params[2],
                params[3],
                params[4],
                params[5]
            )

        response: GeoComResponse[
            tuple[int, int, int, int, int, int]
        ] = self._request(
            5051,
            parsers=(
                int,
                int,
                int,
                int,
                int,
                int
            )
        )

        return response.map_params(make_datetime)

    def set_datetime_new(
        self,
        time: datetime
    ) -> GeoComResponse[None]:
        """
        RPC 5050, ``CSV_SetDateTime2``

        .. versionadded:: GeoComp-VivaTPS

        Sets the date and time on the instrument.

        Parameters
        ----------
        time : datetime
            New date and time to set.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_datetime_new

        """
        return self._request(
            5050,
            [
                time.year, time.month, time.day,
                time.hour, time.minute, time.second
            ]
        )

    def setup_listing(self) -> GeoComResponse[None]:
        """
        RPC 5072, ``CSV_SetupList``

        Prepares listing of the jobs in memory.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        list

        """
        return self._request(
            5072
        )

    def list(self) -> GeoComResponse[tuple[str, str, int, int, str]]:
        """
        RPC 5073, ``CSV_List``

        Gets the next job listing entry.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Job name.
                - `str`: File name (`-01`: job, `-02`: code list).
                - `int`: Unknown.
                - `int`: Unknown.
                - `str`: Unknown.

        See Also
        --------
        setup_listing

        """
        return self._request(
            5073,
            parsers=(
                str,
                str,
                int,
                int,
                str
            )
        )

    def get_maintenance_end(self) -> GeoComResponse[datetime]:
        """
        RPC 5114, ``CSV_GetMaintenanceEnd``

        Gets the date when the software maintenance service ends.

        Returns
        -------
        GeoComResponse
            Params:
                - `datetime`: Software maintenance end date.

        """
        def transform(
            params: tuple[int, Byte, Byte] | None
        ) -> datetime | None:
            if params is None:
                return None

            return datetime(
                params[0],
                int(params[1]),
                int(params[2])
            )

        response = self._request(
            5114,
            parsers=[
                int,
                Byte.parse,
                Byte.parse
            ]
        )

        return response.map_params(transform)



"""
Description
===========

Module: ``geocompy.geo.bap``

Definitions for the GeoCom Basic applications subsystem.

Types
-----

- ``GeoComBAP``

"""
from __future__ import annotations

from ..data import (
    Angle,
    toenum,
    enumparser,
    parsestr,
    parsebool
)
from .gcdata import (
    Program,
    Prism,
    Reflector,
    Target,
    UserProgram,
    ATRMode
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComBAP(GeoComSubsystem):
    """
    Basic applications subsystem of the GeoCom protocol.

    This subsystem contains high-level functions that are also accessible
    through the user interface. The commands combine several subcommands
    for ease of operation.

    """

    def get_last_displayed_error(self) -> GeoComResponse[tuple[int, int]]:
        """
        RPC 17003, ``BAP_GetLastDisplayedError``

        .. versionremoved:: GeoCom-TPS1200

        Retrieves the number of the last displayed system error.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Last error, warning or info number.
                - `int`: Corresponding GSI error number.
            Error codes:
                - ``IVRESULT``: No error was displayed since last call.

        Note
        ----
        The error code will be reset once command is executed. A repeated
        call will be unsuccessful.
        """
        return self._request(
            17003,
            parsers=(
                int,
                int
            )
        )

    def measure_distance_angle(
        self,
        mode: Program | str = Program.DISTANCE
    ) -> GeoComResponse[tuple[Angle, Angle, float, Program]]:
        """
        RPC 17017, ``BAP_MeasDistanceAngle``

        Take an angle and distance measuremnt depending on the distance
        mode.

        Parameters
        ----------
        mode : Program | str, optional
            Distance measurement mode to use, by default
            Program.DISTANCE

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
                - `float`: Slope distance.
                - `MEASUREPRG`: Actual distance mode.
            Info codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy cannot be guaranteed.
                - ``TMC_ANGLE_ACCURACY_GUARANTEE``: Only angle measurement
                  valid, accuracy cannot be guaranteed.
            Warning codes:
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angle measurement
                  valid, accuracy cannot be guaranteed.
                - ``TMC_ANGLE_OK``: Only angle measurement valid.
                - ``TMC_NO_FULL_CORRECTION``: Measurement without full
                  correction.
            Error codes:
                - ``AUT_ANGLE_ERROR``: Angle measurement error.
                - ``AUT_BAD_ENVIRONMENT``: Bad environmental conditions.
                - ``AUT_CALACC``: ATR calibration failed.
                - ``AUT_DETECTOR_ERROR``: Error in target acquisition.
                - ``AUT_DETENT_ERROR``: Positioning not possible.
                - ``AUT_DEV_ERROR``: Error in angle deviation calculation.
                - ``AUT_INCACC``: Position not exactly reached.
                - ``AUT_MOTOR_ERROR``: Motorization error.
                - ``AUT_MULTIPLE_TARGETS``: Multiple targets detected.
                - ``AUT_NO_TARGET``: No target detected.
                - ``AUT_TIMEOUT``: Position not reached.
                - ``BAP_CHANGE_ALL_TO_DIST``: Prism not detected, changed
                  command to ALL.
                - ``TMC_ANGLE_ERROR``: No valid angle measurement.
                - ``TMC_BUSY``: TMC submodule already in use by another
                  subsystem, command not processed.
                - ``TMC_DIST_ERROR``: An error occurred during distance
                  measurement.
                - ``TMC_DIST_PPM``: Wrong PPM setting.
                - ``TMC_SIGNAL_ERROR``: No signal on EDM (only in signal
                  mode).
                - ``ABORT``: Measurement aborted.
                - ``COM_TIMEDOUT``: Communication timeout.
                - ``IVPARAM``: Invalid distance mode.
                - ``SHUT_DOWN``: System stopped.

        """
        _mode = toenum(Program, mode)
        return self._request(
            17017,
            [_mode.value],
            (
                Angle.parse,
                Angle.parse,
                float,
                enumparser(Program)
            )
        )

    def get_target_type(self) -> GeoComResponse[Target]:
        """
        RPC 17022, ``BAP_GetTargetType``

        .. versionadded:: GeoCom-TPS1100-1.05

        Gets the current EDM target type.

        Returns
        -------
        GeoComResponse
            Params:
                - `Target`: Current EMD target type.

        See Also
        --------
        set_target_type
        """
        return self._request(
            17022,
            parsers=enumparser(Target)
        )

    def set_target_type(
        self,
        target: Target | str
    ) -> GeoComResponse[None]:
        """
        RPC 17021, ``BAP_SetTargetType``

        .. versionadded:: GeoCom-TPS1100-1.05

        Sets the EDM target type. The last target type is remembered for
        all EDM modes.

        Parameters
        ----------
        target : Target | str
            New EDM target type to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Target type is not available.

        See Also
        --------
        get_target_type
        set_measurement_program
        """
        _target = toenum(Target, target)
        return self._request(
            17021,
            [_target.value]
        )

    def get_prism_type(self) -> GeoComResponse[Prism]:
        """
        RPC 17009, ``BAP_GetPrismType``

        .. versionadded:: GeoCom-TPS1100-1.05

        Gets the current prism type.

        Returns
        -------
        GeoComResponse
            Params:
                - `Prism`: Current prism type.
            Error codes:
                - ``IVRESULT``: EDM is set to reflectorless mode.

        See Also
        --------
        set_prism_type
        """
        return self._request(
            17009,
            parsers=enumparser(Prism)
        )

    def set_prism_type(
        self,
        prism: Prism | str
    ) -> GeoComResponse[None]:
        """
        RPC 17008, ``BAP_SetPrismType``

        .. versionadded:: GeoCom-TPS1100-1.05

        Sets the prism type. Prism change also overwrites the current
        prism constant.

        Parameters
        ----------
        prism : Prism | str
            New prism type to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Prism type is not available.

        See Also
        --------
        get_prism_type
        """
        _prism = toenum(Prism, prism)
        return self._request(
            17008,
            [_prism.value]
        )

    def get_prism_definition(
        self,
        prism: Prism | str
    ) -> GeoComResponse[tuple[str, float, Reflector]]:
        """
        RPC 17023, ``BAP_GetPrismDef``

        .. versionadded:: GeoCom-TPS1100-1.05

        Gets the definition of the default prism.

        Parameters
        ----------
        prism : Prism | str
            Prism type to query.

        Returns
        -------
        GeoComResponse
            Params:
                - `str`: Name of the prism.
                - `float`: Additive prism constant.
                - `Reflector`: Reflector type.
            Error codes:
                - ``IVPARAM``: Invalid prism type.

        """
        _prism = toenum(Prism, prism)
        return self._request(
            17023,
            [_prism.value],
            (
                parsestr,
                float,
                enumparser(Reflector)
            )
        )

    def set_prism_definition(
        self,
        prism: Prism | str,
        name: str,
        const: float,
        reflector: Reflector | str
    ) -> GeoComResponse[None]:
        """
        RPC 17024, ``BAP_SetPrismDef``

        .. versionadded:: GeoCom-TPS1100-1.05

        .. versionremoved:: GeoCom-TPS1200

        Defines a user prism.

        Parameters
        ----------
        prism : Prism | str
            Type of the new prism. (Can be USER1, 2 and 3.)
        name : str
            Definition name. (Maximum 16 characters. Longer names will be
            truncated.)
        const : float
            Additive prism constant.
        reflector : Reflector | str
            Reflector type.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid prism type.

        """
        _prism = toenum(Prism, prism)
        _reflector = toenum(Reflector, reflector)
        name = f"{name:.16s}"
        return self._request(
            17024,
            [
                _prism.value,
                name,
                const,
                _reflector.value
            ]
        )

    def get_measurement_program(self) -> GeoComResponse[UserProgram]:
        """
        RPC 17018, ``BAP_GetMeasPrg``

        Gets the current measurement program.

        Returns
        -------
        GeoComResponse
            Params:
                - `UserProgram`: Current measurement program.

        See Also
        --------
        set_measurement_program
        """
        return self._request(
            17018,
            parsers=enumparser(UserProgram)
        )

    def set_measurement_program(
        self,
        program: UserProgram | str
    ) -> GeoComResponse[None]:
        """
        RPC 17019, ``BAP_SetMeasPrg``

        Sets a new measurement program.

        Parameters
        ----------
        program : UserProgram | str
            Measurement program to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Measurement program is not available.

        See Also
        --------
        get_measurement_program
        set_target_type
        """
        _program = toenum(UserProgram, program)
        return self._request(
            17019,
            [_program.value]
        )

    def search_target(self) -> GeoComResponse[None]:
        """
        RPC 17020, ``BAP_SearchTarget``

        .. versionadded:: GeoCom-TPS1100-1.04

        Executes target search in the predefined window.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``AUT_BAD_ENVIRONMENT``: Bad environmental conditions.
                - ``AUT_DEV_ERROR``: Error in angle deviation calculation.
                - ``AUT_ANGLE_ACCURACY``: Position not exactly reached.
                - ``AUT_MOTOR_ERROR``: Motorization error.
                - ``AUT_MULTIPLE_TARGETS``: Multiple targets detected.
                - ``AUT_NO_TARGET``: No target detected.
                - ``AUT_TIMEOUT``: Position not reached.
                - ``ABORT``: Measurement aborted.
                - ``FATAL``: Fatal error.

        See Also
        --------
        aut.get_spiral
        aut.set_spiral
        aut.get_search_area
        aut.set_search_area
        """
        return self._request(17020, [0])

    def get_prism_type_name(self) -> GeoComResponse[tuple[Prism, str]]:
        """
        RPC 17031, ``BAP_GetPrismType2``

        .. versionadded:: GeoCom-TPS1200

        Gets the current prism type and name.

        Returns
        -------
        GeoComResponse
            Params:
                - `Prism`: Current prism type.
                - `str`: Prism type name.

        See Also
        --------
        set_prism_type
        set_prism_type_name
        """
        return self._request(
            17031,
            parsers=(enumparser(Prism), parsestr)
        )

    def set_prism_type_name(
        self,
        prism: Prism | str,
        name: str
    ) -> GeoComResponse[None]:
        """
        RPC 17030, ``BAP_SetPrismType2``

        .. versionadded:: GeoCom-TPS1200

        Sets the prism type and name.

        Parameters
        ----------
        prism : Prism | str
            Prism type to set.
        name : str
            Name of the prism type.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Prism type is not available, a user prism
                  is not defined.

        See Also
        --------
        get_prism_type_name
        tmc.set_prism_correction
        """
        _prism = toenum(Prism, prism)
        return self._request(
            17030,
            [_prism.value, name]
        )

    def get_user_prism_definition(
        self,
        name: str
    ) -> GeoComResponse[tuple[float, Reflector, str]]:
        """
        RPC 17033, ``BAP_GetUserPrismDef``

        .. versionadded:: GeoCom-TPS1200

        Gets the definition of a user defined prism.

        Parameters
        ----------
        name : str
            Name of the prism.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Additive prism constant.
                - `Reflector`: Reflector type.
                - `str`: Creator of the prism definition.
            Error codes:
                - ``IVPARAM``: Invalid prism definition.

        See Also
        --------
        get_prism_type
        get_prism_type_name
        get_prism_definition
        set_user_prism_definition
        """
        return self._request(
            17033,
            [name],
            (
                float,
                enumparser(Reflector),
                parsestr
            )
        )

    def set_user_prism_definition(
        self,
        name: str,
        const: float,
        reflector: Reflector | str,
        creator: str
    ) -> GeoComResponse[None]:
        """
        RPC 17032, ``BAP_SetUserPrismDef``

        .. versionadded:: GeoCom-TPS1200

        Defines a new user defined prism.

        Parameters
        ----------
        name : str
            Name of the prism.
        const : float
            Additive prism constant.
        reflector: Reflector | str
            Reflector type.
        creator : str
            Name of the creator.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid prism definition.
                - ``IVRESULT``: Prism definition is not set.

        See Also
        --------
        set_prism_type
        get_prism_definition
        set_user_prism_definition
        """
        _reflector = toenum(Reflector, reflector)
        return self._request(
            17032,
            [name, const, _reflector.value, creator]
        )

    def get_atr_setting(self) -> GeoComResponse[ATRMode]:
        """
        RPC 17034, ``BAP_GetATRSetting``

        .. versionadded:: GeoCom-TPS1200

        Gets the current ATR setting.

        Returns
        -------
        GeoComResponse
            Params:
                - `ATRMode`: Current ATR setting.

        See Also
        --------
        set_atr_setting
        """
        return self._request(
            17034,
            parsers=enumparser(ATRMode)
        )

    def set_atr_setting(
        self,
        mode: ATRMode | str
    ) -> GeoComResponse[None]:
        """
        RPC 17035, ``BAP_SetATRSetting``

        .. versionadded:: GeoCom-TPS1200

        Sets the ATR setting.

        Parameters
        ----------
        mode : ATRMode | str
            ATR setting to activate.

        See Also
        --------
        get_atr_setting
        """
        _mode = toenum(ATRMode, mode)
        return self._request(
            17035,
            [_mode.value]
        )

    def get_reduced_atr_fov_status(self) -> GeoComResponse[bool]:
        """
        RPC 17036, ``BAP_GetRedATRFov``

        .. versionadded:: GeoCom-TPS1200

        Gets the state of the reduced ATR field of view mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Reduced field of view ATR is enabled.

        See Also
        --------
        switch_reduced_atr_fov
        """
        return self._request(
            17036,
            parsers=parsebool
        )

    def switch_reduced_atr_fov(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 17037, ``BAP_SetRedATRFov``

        .. versionadded:: GeoCom-TPS1200

        Sets the state of the reduced ATR field of view mode.

        Parameters
        ----------
        enabled : bool
            Reduced field of view ATR is enabled.

        See Also
        --------
        get_reduced_atr_fov_status
        """
        return self._request(
            17037,
            [enabled]
        )

    def get_precise_atr_status(self) -> GeoComResponse[bool]:
        """
        RPC 17039, ``BAP_GetATRPrecise``

        .. versionadded:: GeoCom-VivaTPS

        Gets the current state of the precise ATR mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Precise ATR mode is enabled.

        See Also
        --------
        switch_precise_atr
        """
        return self._request(
            17039,
            parsers=parsebool
        )

    def switch_precise_atr(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 17040, ``BAP_SetATRPrecise``

        .. versionadded:: GeoCom-VivaTPS

        Sets the state of the precise ATR mode.

        Parameters
        ----------
        enabled : bool
            Precise ATR mode is enabled.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_precise_atr_status
        """
        return self._request(
            17040,
            [enabled]
        )



"""
Description
===========

Module: ``geocompy.geo.cam``

Definitions for the GeoCom Camera subsystem.

Types
-----

- ``GeoComCAM``

"""
from __future__ import annotations

from ..data import (
    Coordinate,
    Vector,
    Angle,
    toenum,
    enumparser,
    parsebool
)
from .gcdata import (
    Camera,
    Compression,
    JPEGQuality,
    Resolution,
    WhiteBalance,
    Zoom
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComCAM(GeoComSubsystem):
    """
    Camera subsystem of the GeoCom protocol.

    This subsystem performs tasks relating to the overview camera and
    (on Nova instruments) the telescope mounted camera

    All functions require a valid GeoCom Imaging license.

    .. versionadded:: GeoCom-VivaTPS
    """

    def set_zoom(
        self,
        zoom: Zoom | str,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23608, ``CAM_SetZoomFactor``

        Sets the specified zoom factor on a camera device.

        Parameters
        ----------
        zoom : Zoom | str
            Zoom level to set.
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        _zoom = toenum(Zoom, zoom)
        _camera = toenum(Camera, camera)
        return self._request(
            23608,
            [_camera.value, _zoom.value]
        )

    def get_zoom(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[Zoom]:
        """
        RPC 23609, ``CAM_GetZoomFactor``

        Sets the current zoom factor on a camera device.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Params:
                - `Zoom`: Current zoom level.
            Error codes:
                - ``NA``: Imaging license not found.

        """
        _camera = toenum(Camera, camera)
        return self._request(
            23609,
            [_camera.value],
            enumparser(Zoom)
        )

    def get_camera_position(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[Coordinate]:
        """
        RPC 23611, ``CAM_GetCamPos``

        Gets the position of the overview camera, relative to the station
        coordinates.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Params:
                - `Coordinate`: Relative coordinates of the
                  camera.
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        tmc.get_station
        get_camera_direction
        """
        def transform(
            params: tuple[float, float, float] | None
        ) -> Coordinate | None:
            if params is None:
                return None

            return Coordinate(
                params[0],
                params[1],
                params[2]
            )

        _camera = toenum(Camera, camera)
        response = self._request(
            23611,
            [_camera.value],
            (
                float,
                float,
                float
            )
        )
        return response.map_params(transform)

    def get_camera_direction(
        self,
        dist: float,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[Vector]:
        """
        RPC 23611, ``CAM_GetCamViewingDir``

        Gets the view vector of the overview camera relative to its
        coordinates. The viewing vector is a 3D vector along the optical
        axis of the camera, with the given slope distance length.

        Parameters
        ----------
        dist : float
            View vector length.
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Params:
                - `Coordinate`: Viewing vector.
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        get_camera_position
        """
        def transform(
            params: tuple[float, float, float] | None
        ) -> Vector | None:
            if params is None:
                return None

            return Vector(
                params[0],
                params[1],
                params[2]
            )

        _camera = toenum(Camera, camera)
        response = self._request(
            23613,
            [_camera.value, dist],
            (
                float,
                float,
                float
            )
        )

        return response.map_params(transform)

    def get_camera_fov(
        self,
        camera: Camera | str = Camera.OVERVIEW,
        zoom: Zoom | str = Zoom.X1
    ) -> GeoComResponse[tuple[Angle, Angle]]:
        """
        RPC 23619, ``CAM_GetCameraFoV``

        Gets field of view of the overview camera for a given zoom level.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW
        zoom : Zoom | str, optional
            Zoom level, by default X1

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal field of view.
                - `Angle`: Vertical field of view.
            Error codes:
                - ``IVPARAM``: Invalid parameter.
                - ``NA``: Imaging license not found.

        """
        _camera = toenum(Camera, camera)
        _zoom = toenum(Zoom, zoom)
        return self._request(
            23619,
            [_camera.value, _zoom.value],
            (
                Angle.parse,
                Angle.parse
            )
        )

    def set_actual_image_name(
        self,
        name: str,
        number: int,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23619, ``CAM_SetActualImageName``

        Sets the name and number of the next image to be taken.

        Parameters
        ----------
        name : str
            Image name.
        number : int
            Image number.
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        take_image
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23622,
            [_camera.value, name, number]
        )

    def take_image(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23623, ``CAM_TakeImage``

        Takes a new image with the selected camera.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``CAM_IMAGE_SAVING_ERROR``: Error while saving, SD card
                  might not be available.
                - ``NA``: Imaging license not found.

        See Also
        --------
        is_camera_ready
        set_camera_properties
        set_actual_image_name
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23623,
            [_camera.value]
        )

    def get_overview_crosshair(self) -> GeoComResponse[tuple[float, float]]:
        """
        RPC 23624, ``CAM_GetActCameraCenter``

        Calculates the position of the optical crosshair on the overview
        camera image, at a previously set distance.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Horizontal position of corsshair on
                  image.
                - `float`: Vertical position of corsshair on
                  image.
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        set_overview_distance
        set_camera_properties
        """
        return self._request(
            23624,
            parsers=(
                float,
                float
            )
        )

    def set_overview_distance(
        self,
        dist: float,
        face1: bool = True
    ) -> GeoComResponse[None]:
        """
        RPC 23625, ``CAM_GetActDistance``

        Sets distance to the current target.

        Parameters
        ----------
        dist : float
            Target distance.
        face1 : float
            Telescope is in face 1 position.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        get_overview_crosshair
        """
        return self._request(
            23625,
            [dist, face1]
        )

    def set_whitebalance(
        self,
        whitebalance: WhiteBalance | str = WhiteBalance.AUTO,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23626, ``CAM_SetWhiteBalanceMode``

        Sets the white balance mode for a camera device.

        Parameters
        ----------
        whitebalance : WhiteBalance | str, optional
            White balance mode, by default WhiteBalance.AUTO
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        take_image
        """
        _wb = toenum(WhiteBalance, whitebalance)
        _camera = toenum(Camera, camera)
        return self._request(
            23626,
            [_camera.value, _wb.value]
        )

    def is_camera_ready(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23627, ``CAM_IsCameraReady``

        Checks if a camera is ready for use.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.
                - ``CAM_NOT_READY``: Camera is turned off, or currently
                  starting up.

        See Also
        --------
        get_camera_power_status
        switch_camera_power
        wait_for_camera_ready
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23627,
            [_camera.value]
        )

    def set_camera_properties(
        self,
        resolution: Resolution | str,
        compression: Compression | str,
        jpegquality: JPEGQuality | str,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23633, ``CAM_SetCameraProperties``

        Sets camera parameters.

        Parameters
        ----------
        resolution : Resolution | str
            Image resolution.
        compression : Compression | str
            Image compression.
        jpegquality : JPEGQuality | str
            JPEG image compression quality.
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        is_camera_ready
        set_actual_image_name
        take_image
        """
        _res = toenum(Resolution, resolution)
        _comp = toenum(Compression, compression)
        _qual = toenum(JPEGQuality, jpegquality)
        _camera = toenum(Camera, camera)
        return self._request(
            23633,
            [_camera.value, _res.value, _comp.value, _qual.value]
        )

    def get_camera_power_status(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[bool]:
        """
        RPC 23636, ``CAM_GetCameraPowerSwitch``

        Gets the current state of the camera.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Camera is powered and active.
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        is_camera_ready
        switch_camera_power
        wait_for_camera_ready
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23636,
            [_camera.value],
            parsebool
        )

    def switch_camera_power(
        self,
        activate: bool,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23637, ``CAM_SetCameraPowerSwitch``

        Sets the state of the camera.

        Parameters
        ----------
        activate : bool
            Power up and activate camera.
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        is_camera_ready
        get_camera_power_status
        wait_for_camera_ready
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23637,
            [_camera.value, activate]
        )

    def wait_for_camera_ready(
        self,
        wait: int = 30,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23638, ``CAM_WaitForCameraReady``

        Waits for the camera device to become ready for use.

        Parameters
        ----------
        wait : int, optional
            Time to wait for the camera to come online [s].
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.
                - ``TIMEOUT``: Camera did not become usable within the
                  specified time.

        See Also
        --------
        is_camera_ready
        get_camera_power_status
        switch_camera_power
        """
        _camera = toenum(Camera, camera)
        return self._request(
            23638,
            [_camera.value, int(wait * 1000)]
        )

    def set_autofocus_position(
        self,
        position: int
    ) -> GeoComResponse[None]:
        """
        RPC 23645, ``CAM_AF_SetMotorPosition``

        Sets the autofocus motor to a specific position.

        Parameters
        ----------
        position : int
            Autofocus motor position.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        get_autofocus_position
        """
        return self._request(
            23645,
            [position]
        )

    def get_autofocus_position(self) -> GeoComResponse[int]:
        """
        RPC 23644, ``CAM_AF_GetMotorPosition``

        Gets the current position of the autofocus motor.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Autofocus motor position.
            Error codes:
                - ``NA``: Imaging license not found.

        See Also
        --------
        set_autofocus_position
        """
        return self._request(
            23644,
            parsers=int
        )

    def switch_continuous_autofocus(
        self,
        start: bool
    ) -> GeoComResponse[None]:
        """
        RPC 23669, ``CAM_AF_ContinuousAutofocus``

        Starts or stops the continuous autofocus.

        Parameters
        ----------
        start : bool
            Start the continuous autofocus.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(
            23669,
            [start]
        )

    def set_focus_to_distance(
        self,
        dist: float
    ) -> GeoComResponse[None]:
        """
        RPC 23652, ``CAM_AF_PositFocusMotorToDist``

        Sets the autofocus motor to the specified distance.

        Parameters
        ----------
        dist : float
            Distance to focus to.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(
            23652,
            [dist]
        )

    def set_focus_to_infinity(self) -> GeoComResponse[None]:
        """
        RPC 23677, ``CAM_AF_PositFocusMotorToInfinity``

        Sets the autofocus motor to focus to infinity.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(23677)

    def set_focus_to_target(self) -> GeoComResponse[None]:
        """
        RPC 23662, ``CAM_AF_SingleShotAutofocus``

        Focuses current target.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(23662)

    def set_focus_to_target_contrast(
        self,
        steps: int
    ) -> GeoComResponse[None]:
        """
        RPC 23663, ``CAM_AF_FocusContrastAroundCurrent``

        Focuses current target by contrast around target.

        Parameters
        ----------
        steps : int
            Focus iteration steps.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(
            23663,
            [steps]
        )

    def get_sensor_size(
        self,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[tuple[float, float]]:
        """
        RPC 23668, ``CAM_GetChipWindowSize``

        Gets the size of the camera chip.

        Parameters
        ----------
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Sensor width.
                - `float`: Sensor height.
            Error codes:
                - ``NA``: Imaging license not found.

        """
        _camera = toenum(Camera, camera)
        return self._request(
            23668,
            [_camera.value],
            (
                float,
                float
            )
        )

    def get_telescopic_crosshair(self) -> GeoComResponse[tuple[int, int]]:
        """
        RPC 23671, ``CAM_OAC_GetCrossHairPos``

        Gets the position of the crosshair in the actual camera resolution.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Horizontal position.
                - `int`: Vertical position.
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(
            23671,
            parsers=(
                int,
                int
            )
        )

    def get_overview_interior_orientation(
        self,
        calibrated: bool = True
    ) -> GeoComResponse[tuple[float, float, float, float]]:
        """
        RPC 23602, ``CAM_OVC_ReadInterOrient``

        Gets the interior orientation parameters of the camera.

        Parameters
        ----------
        calibrated : bool, optional
            Use calibrated data.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Horizontal position of principal point.
                - `float`: Vertical position of principal point.
                - `float`: Focus length.
                - `float`: Pixel size.
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(
            23602,
            [calibrated],
            (
                float,
                float,
                float,
                float
            )
        )

    def get_overview_exterior_orientation(
        self,
        calibrated: bool = True
    ) -> GeoComResponse[tuple[Coordinate, Angle, Angle, Angle]]:
        """
        RPC 23603, ``CAM_OVC_ReadExterOrient``

        Gets the exterior orientation parameters of the camera.

        Parameters
        ----------
        calibrated : bool, optional
            Use calibrated data.

        Returns
        -------
        GeoComResponse
            Params:
                - `Coordinate`: Camera coordinates.
                - `Angle`: Azimut angle.
                - `Angle`: Zenith angle.
                - `Angle`: Tilt angle.
            Error codes:
                - ``NA``: Imaging license not found.

        """
        def transform(
            params: tuple[float, float, float, Angle, Angle, Angle] | None
        ) -> tuple[Coordinate, Angle, Angle, Angle] | None:
            if params is None:
                return None
            return (
                Coordinate(
                    params[0],
                    params[1],
                    params[2]
                ),
                params[3],
                params[4],
                params[5]
            )

        response = self._request(
            23603,
            [calibrated],
            (
                float,
                float,
                float,
                Angle.parse,
                Angle.parse,
                Angle.parse
            )
        )

        return response.map_params(transform)

    def start_remote_video(
        self,
        fps: int,
        bitrate: int,
        camera: Camera | str = Camera.OVERVIEW
    ) -> GeoComResponse[None]:
        """
        RPC 23675, ``CAM_StartRemoteVideo``

        Starts a remote video stream that can be watched when connected
        wirelessly to the instrument. Networkstrea:
        ``rtsp://192.168.254.3/TSCame``.

        Parameters
        ----------
        fps : int
            Frame rate 3/5/10 [Hz].
        bitrate : int
            Video bit rate in [100; 6144] range [kbps].
        camera : Camera | str, optional
            Camera device, by default OVERVIEW

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        _camera = toenum(Camera, camera)
        return self._request(
            23675,
            [_camera.value, fps, bitrate]
        )

    def stop_remote_video(self) -> GeoComResponse[None]:
        """
        RPC 23676, ``CAM_StopRemoteVideo``

        Stops the remote video stream.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Imaging license not found.

        """
        return self._request(23676)



"""
Description
===========

Module: ``geocompy.gsi``

The ``gsi`` package provides wrapper methods for GSI Online commands,
supported by many totalstations (TPS) and digital levels (NA/DNA).


Subpackages
-----------

``geocompy.gsi.dna``
    GSI Online digital level commands.

Submodules
----------

- ``geocompy.gsi.gsidata``
- ``geocompy.gsi.gsitypes``
"""



"""
Description
===========

Module: ``geocompy.geo.kdm``

Definitions for the GeoCom Keyboard display unit subsystem.

Types
-----

- ``GeoComKDM``

"""
from __future__ import annotations

from ..data import parsebool
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComKDM(GeoComSubsystem):
    """
    Keyboard display unit subsystem of the GeoCom protocol.

    This subsystem controls the keyboard and display functions.

    """

    def switch_display_power(
        self,
        alwayson: bool
    ) -> GeoComResponse[None]:
        """
        RPC 23107, ``KDM_SetLcdPower``

        Sets the status of the diplay power.

        Parameters
        ----------
        alwayson : bool
            Keep display turned on, do not go into screensaver mode.

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            23107,
            [alwayson]
        )

    def get_display_power_status(self) -> GeoComResponse[bool]:
        """
        RPC 23108, ``KDM_GetLcdPower``

        Gets the current status of the diplay power.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Keep display turned on, do not go
                  into screensaver mode.

        """
        return self._request(
            23108,
            parsers=parsebool
        )



from typing import Callable, Any, Iterable
import re

import pytest

from geocompy.geo import GeoCom
from geocompy.communication import Connection
from geocompy.data import Byte

from helpers import faulty_parser, FaultyConnection


@pytest.fixture
def instrument() -> GeoCom:
    return GeoCom(DummyGeoComConnection())


class DummyGeoComConnection(Connection):
    _RESP = re.compile(
        r"^%R1P,"
        r"(?P<comrc>\d+),"
        r"(?P<tr>\d+):"
        r"(?P<rc>\d+)"
        r"(?:,(?P<params>.*))?$"
    )

    _CMD = re.compile(
        r"^%R1Q,"
        r"(?P<rpc>\d+)"
        r"(?P<trid>,\d+)?:"
        r"(?:(?P<params>.*))?$"
    )

    def send(self, message: str) -> None:
        return

    def receive(self) -> str:
        return ""

    def exchange(self, cmd: str) -> str:
        if not self._CMD.match(cmd):
            return "%R1P,0,0:2"

        head, _ = cmd.split(":")
        match head.split(","):
            case [_, _, trid_str]:
                pass
            case _:
                trid_str = "0"

        trid = int(trid_str)

        if re.match(r"%R1Q,5008,\d+:", cmd):
            return f"%R1P,0,{trid}:0,1996,'07','19','10','13','2f'"

        return f"%R1P,0,{trid}:0"

    def close(self) -> None:
        pass

    def reset(self) -> None:
        pass

    def is_open(self) -> bool:
        return True


class TestGeoCom:
    def test_init(self) -> None:
        conn_bad = FaultyConnection()
        with pytest.raises(ConnectionError):
            GeoCom(conn_bad, retry=1)

        conn_good = DummyGeoComConnection()
        instrument = GeoCom(conn_good)
        assert instrument._precision == 15

    def test_parse_response(self, instrument: GeoCom) -> None:
        cmd = "%R1Q,5008,0:"
        answer = "%R1P,0,0:0,1996,'07','19','10','13','2f'"
        parsers: Iterable[Callable[[str], Any]] = (
            int,
            Byte.parse,
            Byte.parse,
            Byte.parse,
            Byte.parse,
            Byte.parse
        )
        response = instrument.parse_response(
            cmd,
            answer,
            parsers
        )
        assert response.params is not None
        assert response.params[0] == 1996

        response = instrument.parse_response(
            cmd,
            "%R1P,1,0:",
            parsers
        )
        assert response.params is None

        parsers_faulty = (
            faulty_parser,
            Byte.parse,
            Byte.parse,
            Byte.parse,
            Byte.parse,
            Byte.parse
        )
        response = instrument.parse_response(
            cmd,
            answer,
            parsers_faulty
        )
        assert response.params is None

    def test_request(self, instrument: GeoCom) -> None:
        response = instrument.request(
            5008,
            parsers=(
                int,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse,
                Byte.parse
            )
        )
        assert response.params is not None
        assert response.params[0] == 1996

        response = instrument.request(
            1,
            (1, 2.0)
        )
        assert re.match(r"%R1Q,1,\d+:1,2.0", response.cmd)
        assert re.match(r"%R1P,0,\d+:0", response.response)



"""
Description
===========

Module: ``geocompy.geo.com``

Definitions for the GeoCom Communication subsystem.

Types
-----

- ``GeoComCOM``

"""
from __future__ import annotations

from ..data import (
    toenum,
    parsebool
)
from .gcdata import (
    Shutdown,
    Startup
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComCOM(GeoComSubsystem):
    """
    Communication subsystem of the GeoCom protocol.

    This subsystem contains functions relevant to the communication
    with the instrument.

    """

    def get_geocom_version(self) -> GeoComResponse[tuple[int, int, int]]:
        """
        RPC 110, ``COM_GetSWVersion``

        Gets the version of the installed GeoCom release.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Release number.
                - `int`: Version number.
                - `int`: Subversion number.

        See Also
        --------
        csv.get_firmware_version
        """
        return self._request(
            110,
            parsers=(int, int, int)
        )

    def set_send_delay(
        self,
        delay: float
    ) -> GeoComResponse[None]:
        """
        RPC 109, ``COM_SetSendDelay``

        .. versionremoved:: GeoCom-TPS1200

        Sets response delay on the instrument.

        Parameters
        ----------
        delay : float
            Response delay [s].

        Returns
        -------
        GeoComResponse
        """
        return self._request(109, [int(delay * 1000)])

    def switch_to_local(self) -> GeoComResponse[None]:
        """
        RPC 1, ``COM_Local``

        Switches instrument to local mode, exiting the online mode.

        Returns
        -------
        GeoComResponse

        Warning
        -------
        Once the instrument is switched to local mode, all further RPCs
        will be ignored, until the online mode is manually activated again.
        """
        return self._request(1)

    def switch_on(
        self,
        mode: Startup | str = Startup.REMOTE
    ) -> GeoComResponse[None]:
        """
        RPC 111, ``COM_SwitchOnTPS``

        Switches on the instrument.

        Parameters
        ----------
        mode : Startup | str, optional
            Desired startup mode, by default Startup.REMOTE

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: Instrument is already on.

        Notes
        -----
        The instrument can be switched on with any command, or even just
        a single character.

        See Also
        --------
        switch_off
        """
        _mode = toenum(Startup, mode)
        return self._request(
            111,
            [_mode.value]
        )

    def switch_off(
        self,
        mode: Shutdown | str = Shutdown.SHUTDOWN
    ) -> GeoComResponse[None]:
        """
        RPC 112, ``COM_SwitchOffTPS``

        Switches off the instrument.

        Parameters
        ----------
        mode : Shutdown | str, optional
            Desired stop mode, by default Shutdown.SHUTDOWN

        Returns
        -------
        GeoComResponse

        See Also
        --------
        switch_on
        """
        _mode = toenum(Shutdown, mode)
        return self._request(
            112,
            [_mode.value]
        )

    def nullprocess(self) -> GeoComResponse[None]:
        """
        RPC 0, ``COM_NullProc``

        Tests connection by executing the null process.

        """
        return self._request(0)

    def switch_signoff(
        self,
        enable: bool
    ) -> GeoComResponse[None]:
        """
        RPC 115, ``COM_EnableSignOff``

        .. versionremoved:: GeoCom-TPS1200

        Enables or disables the signoff message upon operation mode
        changes.

        Parameters
        ----------
        enable : bool

        Returns
        -------
        GeoComResponse

        Note
        ----
        This setting is not persistent between sessions.
        """
        return self._request(115, [enable])

    def get_binary_available(self) -> GeoComResponse[bool]:
        """
        RPC 113, ``COM_GetBinaryAvailable``

        Checks if the instrument supports binary communication.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Availability of binary mode.

        See Also
        --------
        set_binary_available
        """
        return self._request(
            113,
            parsers=parsebool
        )

    def set_binary_available(
        self,
        enable: bool
    ) -> GeoComResponse[None]:
        """
        RPC 114, ``COM_SetBinaryAvailable``

        Enables or disables binary communication with the instrument.

        Parameters
        ----------
        enable : bool
            Enable or disable binary communication.

        See Also
        --------
        get_binary_available
        """
        return self._request(
            114,
            [enable]
        )
